#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Utility functions for scaling.

Author: Andrew Lee
"""

from copy import deepcopy
import math
import sys

import json

from pyomo.environ import (
    Binary,
    Block,
    Boolean,
    Constraint,
    NegativeIntegers,
    NegativeReals,
    NonNegativeIntegers,
    NonNegativeReals,
    NonPositiveIntegers,
    NonPositiveReals,
    PositiveIntegers,
    PositiveReals,
    Suffix,
    value,
    Var,
)
from pyomo.core.base.block import BlockData
from pyomo.core.base.var import VarData
from pyomo.core.base.param import ParamData
from pyomo.core import expr as EXPR
from pyomo.common.numeric_types import native_types
from pyomo.core.base.units_container import _PyomoUnit

from idaes.core.util.exceptions import BurntToast
import idaes.logger as idaeslog


_log = idaeslog.getLogger(__name__)

TAB = " " * 4


def get_scaling_suffix(component):
    """
    Get scaling suffix for component.

    If component is not a Block, gets scaling suffix from parent block.
    Creates a new suffix if one is not found.

    Args:
        component: component to get suffix for

    Returns:
        Pyomo scaling Suffix

    Raises:
        TypeError is component is an IndexedBlock
    """
    if isinstance(component, BlockData):
        blk = component
    elif isinstance(component, Block):
        raise TypeError(
            "IndexedBlocks cannot have scaling factors attached to them. "
            "Please assign scaling factors to the elements of the IndexedBlock."
        )
    else:
        blk = component.parent_block()

    try:
        sfx = blk.scaling_factor
    except AttributeError:
        # No existing suffix, create one
        _log.debug(f"Created new scaling suffix for {blk.name}")
        sfx = blk.scaling_factor = Suffix(direction=Suffix.EXPORT)

    return sfx


def scaling_factors_to_dict(blk_or_suffix, descend_into: bool = True):
    """
    Write scaling suffixes to a serializable json dict.

    Component objects are replaced by their local names so they can be
    serialized.

    Args:
        blk_or_suffix: Pyomo Block or Suffix object to covert to dict
        descend_into: for Blocks, whether to descend into any child blocks

    Returns
        dict containing scaling factors indexed by component names

    Raises:
        TypeError if blk_or_suffix is not an instance of Block or Suffix

    """
    # First, determine what type of component we have
    if isinstance(blk_or_suffix, Suffix):
        # Suffix
        sdict = _suffix_to_dict(blk_or_suffix)
        blk = blk_or_suffix.parent_block()
    elif isinstance(blk_or_suffix, BlockData):
        # Scalar block or element of indexed block
        sdict = _collect_block_suffixes(blk_or_suffix, descend_into=descend_into)
        blk = blk_or_suffix
    elif isinstance(blk_or_suffix, Block):
        # Indexed block
        blk = blk_or_suffix
        sdict = {}
        sdict["block_datas"] = {}
        for bd in blk_or_suffix.values():
            sdict["block_datas"][bd.name] = _collect_block_suffixes(
                bd, descend_into=descend_into
            )
    else:
        # Not a block or suffix
        raise TypeError(
            f"{blk_or_suffix.name} is not an instance of a Block of Suffix."
        )

    # Attach block name for future verification
    sdict["block_name"] = blk.name

    return sdict


def scaling_factors_from_dict(
    blk_or_suffix, json_dict: dict, overwrite: bool = False, verify_names: bool = True
):
    """
    Set scaling factors based on values in a serializable json dict.

    This method expects components to be referenced by their local names.

    Args:
        blk_or_suffix: Pyomo Block or Suffix object to set scaling factors on
        json_dict: dict of scaling factors to load into model
        overwrite: (bool) whether to overwrite existing scaling factors or not
        verify_names: (bool) whether to verify that all names in dict exist on block

    Returns
        None

    Raises:
        TypeError if blk_or_suffix is not an instance of Block or Suffix

    """
    # First, copy json_dict so we do not mutate original
    sdict = deepcopy(json_dict)
    # Pop block name for verification
    block_name = sdict.pop("block_name")

    # Next, determine what type of component we have
    if isinstance(blk_or_suffix, Suffix):
        # Suffix
        if verify_names and block_name != blk_or_suffix.parent_block().name:
            raise ValueError(
                f"Name of parent block ({blk_or_suffix.parent_block().name}) does "
                f"not match that recorded in json_dict ({block_name})"
            )
        _suffix_from_dict(
            blk_or_suffix, sdict, overwrite=overwrite, verify_names=verify_names
        )
    elif isinstance(blk_or_suffix, BlockData):
        # Scalar block or element of indexed block
        if verify_names and block_name != blk_or_suffix.name:
            raise ValueError(
                f"Block name ({blk_or_suffix.name}) does "
                f"not match that recorded in json_dict ({block_name})"
            )
        _set_block_suffixes_from_dict(
            blk_or_suffix, sdict, overwrite=overwrite, verify_names=verify_names
        )
    elif isinstance(blk_or_suffix, Block):
        # Indexed block
        if verify_names and block_name != blk_or_suffix.name:
            raise ValueError(
                f"Block name ({blk_or_suffix.name}) does "
                f"not match that recorded in json_dict ({block_name})"
            )
        for bd_name, bd_dict in sdict["block_datas"].items():
            bd = blk_or_suffix.parent_block().find_component(bd_name)
            _set_block_suffixes_from_dict(
                bd, bd_dict, overwrite=overwrite, verify_names=verify_names
            )
    else:
        # Not a block or suffix
        raise TypeError(
            f"{blk_or_suffix.name} is not an instance of a Block of Suffix."
        )


def scaling_factors_to_json_file(blk_or_suffix, filename: str):
    """
    Serialize scaling factors to file in json format.

    Args:
        blk_of_suffix: Block or Suffix to save scaling factors for
        filename: name of file to write to as string

    Returns:
        None

    Raises:
        TypeError if blk_or_suffix is not an instance of Block or Suffix
    """
    with open(filename, "w") as fd:
        json.dump(scaling_factors_to_dict(blk_or_suffix), fd, indent=3)


def scaling_factors_from_json_file(
    blk_or_suffix, filename: str, overwrite: bool = False, verify_names: bool = True
):
    """
    Load scaling factors from json file.

    Args:
        blk_of_suffix: Block or Suffix to load scaling factors for
        filename: name of file to load as string
        overwrite: (bool) whether to overwrite existing scaling factors or not
        verify_names: (bool) whether to verify that all names in dict exist on block

    Returns:
        None

    Raises:
        TypeError if blk_or_suffix is not an instance of Block or Suffix
    """
    with open(filename, "r") as f:
        scaling_factors_from_dict(
            blk_or_suffix, json.load(f), overwrite=overwrite, verify_names=verify_names
        )
    f.close()


def _collect_block_suffixes(block_data, descend_into=True):
    suffix = get_scaling_suffix(block_data)
    sdict = _suffix_to_dict(suffix)

    if descend_into:
        sdict["subblock_suffixes"] = {}
        for sb in block_data.component_data_objects(Block, descend_into=False):
            sdict["subblock_suffixes"][sb.local_name] = _collect_block_suffixes(
                sb, descend_into
            )

    return sdict


def _set_block_suffixes_from_dict(
    block_data, json_dict, verify_names=True, overwrite=False
):
    # First, copy dict so we can take it apart
    sdict = deepcopy(json_dict)

    # Pop any subblock suffixes
    sb_dict = sdict.pop("subblock_suffixes", None)

    # Set local suffix values
    suffix = get_scaling_suffix(block_data)
    _suffix_from_dict(suffix, sdict, verify_names=verify_names, overwrite=overwrite)

    if sb_dict is not None:
        # Get each subblock and apply function recursively
        for sb, sb_dict_value in sb_dict.items():
            subblock = block_data.find_component(sb)

            if subblock is not None:
                _set_block_suffixes_from_dict(
                    subblock,
                    sb_dict_value,
                    verify_names=verify_names,
                    overwrite=overwrite,
                )
            elif verify_names:
                raise AttributeError(
                    f"Block {block_data.name} does not have a subblock named {sb}."
                )


def _suffix_to_dict(suffix):
    sdict = {}

    for k, v in suffix.items():
        # Record components by their local name so we can use
        # find_Component to retrieve them later
        sdict[k.local_name] = v

    return sdict


def _suffix_from_dict(suffix, json_dict, verify_names=True, overwrite=False):
    parent_block = suffix.parent_block()

    for k, v in json_dict.items():
        # Safety catch in case this gets left in by other functions
        if k == "parent_name":
            continue

        comp = parent_block.find_component(k)
        if comp is not None:
            if overwrite or comp not in suffix:
                suffix[comp] = v
        elif verify_names:
            raise ValueError(
                f"Could not find component {k} on block {parent_block.name}."
            )


def get_scaling_factor(component):
    """
    Get scaling factor for component.

    Args:
        component: component to get scaling factor for

    Returns:
        float scaling factor

    Raises:
        TypeError if component is a Block
    """
    if isinstance(component, (Block, BlockData)):
        raise TypeError("Blocks cannot have scaling factors.")

    try:
        sfx = get_scaling_suffix(component)
        return sfx[component]
    except (AttributeError, KeyError):
        # No scaling factor found, return None
        return None


def set_scaling_factor(component, scaling_factor: float, overwrite: bool = False):
    """
    Set scaling factor for component.

    Scaling factors must be positive, non-zero floats.

    Args:
        component: component to set scaling factor for
        scaling_factor: scaling factor to assign
        overwrite: (bool) whether to overwrite existing scaling factor

    Returns:
        None

    Raises:
        ValueError is scaling_factor is 0 or negative
    """
    # Cast to float to catch any garbage inputs
    scaling_factor = float(scaling_factor)

    # Check for negative or 0 scaling factors
    if scaling_factor < 0:
        raise ValueError(
            f"scaling factor for {component.name} is negative ({scaling_factor}). "
            "Scaling factors must be strictly positive."
        )
    elif scaling_factor == 0:
        raise ValueError(
            f"scaling factor for {component.name} is zero. "
            "Scaling factors must be strictly positive."
        )

    # Get suffix and assign scaling factor
    sfx = get_scaling_suffix(component)

    if not overwrite and component in sfx:
        _log.debug(
            f"Existing scaling factor for {component.name} found and overwrite=False. "
            "Scaling factor unchanged."
        )
    else:
        sfx[component] = scaling_factor


def del_scaling_factor(component, delete_empty_suffix: bool = False):
    """
    Delete scaling factor for component.

    Args:
        component: component to delete scaling factor for
        delete_empty_suffix: (bool) whether to delete scaling Suffix if it is
          empty after deletion.
    """
    # Get suffix
    parent = component.parent_block()
    sfx = get_scaling_suffix(parent)

    # Delete entry for component if it exists
    # Pyomo handles case where value does not exist in suffix with a no-op
    sfx.clear_value(component)

    if delete_empty_suffix:
        # Check if Suffix is empty (i.e. length 0)
        if len(sfx) == 0:
            # If so, delete suffix from parent block of component
            _log.debug(f"Deleting empty scaling suffix from {parent.name}")
            parent.del_component(sfx)


def report_scaling_factors(
    blk: Block, ctype=None, descend_into: bool = False, stream=None
):
    """
    Write the scaling factors for all components in a Block to a stream.

    Args:
        blk: Block to get scaling factors from.
        ctype: None, Var or Constraint. Type of component to show scaling factors for
          (if None, shows both Vars and Constraints).
        descend_into: whether to show scaling factors for components in sub-blocks.
        stream: StringIO object to write results to. If not provided, writes to stdout.

    Raises:
        TypeError if blk is not a Pyomo Block.
        ValueError is ctype is not None, Var or Constraint.
    """
    if stream is None:
        stream = sys.stdout

    if ctype not in [None, Var, Constraint]:
        raise ValueError(
            f"report_scaling_factors only supports None, Var or Constraint for argument ctype: "
            f"received {ctype}."
        )

    if not isinstance(blk, (Block, BlockData)):
        raise TypeError(
            "report_scaling_factors: blk must be an instance of a Pyomo Block."
        )

    stream.write(f"Scaling Factors for {blk.name}\n")

    # We will report Vars and Constraint is separate sections for clarity - iterate separately
    if ctype != Constraint:
        # Collect Vars
        vdict = {}
        for blkdata in blk.values():
            for vardata in blkdata.component_data_objects(
                Var, descend_into=descend_into
            ):
                val = vardata.value
                sf = get_scaling_factor(vardata)

                if sf is not None:
                    sfstr = "{:.3E}".format(sf)
                else:
                    sfstr = "None     "

                if val is not None:
                    vstr = "{:.3E}".format(val)
                    if sf is not None:
                        sval = "{:.3E}".format(value(vardata * sf))
                    else:
                        sval = vstr
                else:
                    vstr = "None     "
                    sval = "None"

                vdict[vardata.name] = (sfstr, vstr, sval)

        # Write Var section - skip if no Vars
        if len(vdict) > 0:
            # Get longest var name
            header = "Variable"
            maxname = len(max(vdict.keys(), key=len))
            if maxname < len(header):
                maxname = len(header)

            stream.write(
                f"\n{header}{' '*(maxname-len(header))}{TAB}Scaling Factor{TAB}Value{' '*4}{TAB}Scaled Value\n"
            )

            for n, i in vdict.items():
                # Pad name to length
                stream.write(
                    f"{n + ' '*(maxname-len(n))}{TAB}{i[0]}{' '*5}{TAB}{i[1]}{TAB}{i[2]}\n"
                )

    if ctype != Var:
        # Collect Constraints
        for blkdata in blk.values():
            cdict = {}
            for condata in blkdata.component_data_objects(
                Constraint, descend_into=descend_into
            ):
                sf = get_scaling_factor(condata)

                if sf is not None:
                    sfstr = "{:.3E}".format(sf)
                else:
                    sfstr = "None"

                cdict[condata.name] = sfstr

        # Write Constraint section - skip if no Constraints
        if len(cdict) > 0:
            # Get longest con name
            header = "Constraint"
            maxname = len(max(cdict.keys(), key=len))
            if maxname < len(header):
                maxname = len(header)

            stream.write(
                f"\n{header}{' ' * (maxname - len(header))}{TAB}Scaling Factor\n"
            )

            for n, i in cdict.items():
                # Pad name to length
                stream.write(f"{n + ' ' * (maxname - len(n))}{TAB}{i}\n")


def get_nominal_value(component):
    """
    Get the signed nominal value for a VarData or ParamData component.

    For fixed Vars and Params, the current value of the component will be returned.

    For unfixed Vars, the nominal value is determined using the assigned scaling factor
    and the sign determined based on the bounds and domain of the variable (defaulting to
    positive). If no scaling factor is set, then the current value will be used if set,
    otherwise the absolute nominal value will be equal to 1.

    Args:
        component: component to determine nominal value for

    Returns:
        signed float with nominal value

    Raises:
        TypeError if component is not instance of VarData or ParamData
    """
    # Determine if Var or Param
    if isinstance(component, VarData):
        if component.fixed:
            # Nominal value of a fixed Var is its value
            return value(component)

        # Get scaling factor for Var
        sf = get_scaling_factor(component)
        if sf is None:
            # No scaling factor - see if Var has a value
            if component.value is not None:
                # If it has a value, use that as the nominal value
                # As we are using the actual value, do not need to determine sign
                return value(component)
            else:
                # Otherwise assign a nominal value of 1
                sf = 1

        # Try to determine expected sign of node
        ub = component.ub
        lb = component.lb
        domain = component.domain

        # To avoid NoneType errors, assign dummy values in place of None
        if ub is None:
            # No upper bound, take a positive value
            ub = 1000
        if lb is None:
            # No lower bound, take a negative value
            lb = -1000

        if lb >= 0 or domain in [
            NonNegativeReals,
            PositiveReals,
            PositiveIntegers,
            NonNegativeIntegers,
            Boolean,
            Binary,
        ]:
            # Strictly positive
            sign = 1
        elif ub <= 0 or domain in [
            NegativeReals,
            NonPositiveReals,
            NegativeIntegers,
            NonPositiveIntegers,
        ]:
            # Strictly negative
            sign = -1
        else:
            # Unbounded, see if there is a current value
            # Assume positive until proven otherwise
            sign = 1
            if component.value is not None:
                val = value(component)
                if val < 0:
                    # Assigned negative value, assume value will remain negative
                    sign = -1

        return sign / sf

    elif isinstance(component, ParamData):
        # Nominal value of a parameter is always its value
        return value(component)
    else:
        # Not a Var or Param - invalid component type
        raise TypeError(
            f"get_nominal_value - {component.name} is not an instance of a Var or Param."
        )


class NominalValueExtractionVisitor(EXPR.StreamBasedExpressionVisitor):
    """
    Expression walker for collecting scaling factors in an expression and determining the
    expected value of the expression using the scaling factors as nominal inputs.

    By default, the get_nominal_value method is used to determine the nominal value for
    all Vars and Params in the expression, however this can be changed by setting the
    nominal_value_callback argument.

    Returns a list of expected values for each additive term in the expression.

    In order to properly assess the expected value of terms within functions, the sign
    of each term is maintained throughout thus returned values may be negative. Functions
    using this walker should handle these appropriately.
    """

    def __init__(self, nominal_value_callback=get_nominal_value):
        """
        Visitor class used to determine nominal values of all terms in an expression based on
        scaling factors assigned to the associated variables. Do not use this class directly.

        Args:
            nominal_value_callback - method to use to get nominal value of root nodes.

        Notes
        -----
        This class inherits from the :class:`StreamBasedExpressionVisitor` to implement
        a walker that returns the nominal value corresponding to all additive terms in an
        expression.
        There are class attributes (dicts) that map the expression node type to the
        particular method that should be called to return the nominal value of the node based
        on the nominal value of its child arguments. This map is used in exitNode.
        """
        super().__init__()

        self._nominal_value_callback = nominal_value_callback

    def _get_magnitude_base_type(self, node):
        try:
            return [self._nominal_value_callback(node)]
        except TypeError:
            # Not a Var or Param - something went wrong
            raise BurntToast(
                "NominalValueExtractionVisitor found root node that was not a Var or Param. "
                "This should never happen - please contact the developers with this bug."
            )

    def _get_nominal_value_for_sum_subexpression(self, child_nominal_values):
        return sum(i for i in child_nominal_values)

    def _get_nominal_value_for_sum(self, node, child_nominal_values):
        # For sums, collect all child values into a list
        mag = []
        for i in child_nominal_values:
            for j in i:
                mag.append(j)
        return mag

    def _get_nominal_value_for_product(self, node, child_nominal_values):
        mag = []
        for i in child_nominal_values[0]:
            for j in child_nominal_values[1]:
                mag.append(i * j)
        return mag

    def _get_nominal_value_for_division(self, node, child_nominal_values):
        numerator = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[0]
        )
        denominator = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[1]
        )
        if denominator == 0:
            # Assign a nominal value of 1 so that we can continue
            denominator = 1
            # Log a warning for the user
            _log.warning(
                "Nominal value of 0 found in denominator of division expression. "
                "Assigning a value of 1. You should check you scaling factors and models to "
                "ensure there are no values of 0 that can appear in these functions."
            )
        return [numerator / denominator]

    def _get_nominal_value_for_power(self, node, child_nominal_values):
        # Use the absolute value of the base term to avoid possible complex numbers
        base = abs(
            self._get_nominal_value_for_sum_subexpression(child_nominal_values[0])
        )
        exponent = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[1]
        )

        return [base**exponent]

    def _get_nominal_value_single_child(self, node, child_nominal_values):
        return child_nominal_values[0]

    def _get_nominal_value_abs(self, node, child_nominal_values):
        return [abs(i) for i in child_nominal_values[0]]

    def _get_nominal_value_negation(self, node, child_nominal_values):
        return [-i for i in child_nominal_values[0]]

    def _get_nominal_value_for_unary_function(self, node, child_nominal_values):
        func_name = node.getname()
        func_nominal = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[0]
        )
        func = getattr(math, func_name)
        try:
            return [func(func_nominal)]
        except ValueError:
            raise ValueError(
                f"Evaluation error occurred when getting nominal value in {func_name} "
                f"expression with input {func_nominal}. You should check you scaling factors "
                f"and model to address any numerical issues or scale this constraint manually."
            )

    def _get_nominal_value_expr_if(self, node, child_nominal_values):
        return child_nominal_values[1] + child_nominal_values[2]

    def _get_nominal_value_external_function(self, node, child_nominal_values):
        # First, need to get expected magnitudes of input terms, which may be sub-expressions
        input_mag = [
            self._get_nominal_value_for_sum_subexpression(i)
            for i in child_nominal_values
        ]

        # Next, create a copy of the external function with expected magnitudes as inputs
        newfunc = node.create_node_with_local_data(input_mag)

        # Evaluate new function and return the absolute value
        return [value(newfunc)]

    node_type_method_map = {
        EXPR.EqualityExpression: _get_nominal_value_for_sum,
        EXPR.InequalityExpression: _get_nominal_value_for_sum,
        EXPR.RangedExpression: _get_nominal_value_for_sum,
        EXPR.SumExpression: _get_nominal_value_for_sum,
        EXPR.NPV_SumExpression: _get_nominal_value_for_sum,
        EXPR.ProductExpression: _get_nominal_value_for_product,
        EXPR.MonomialTermExpression: _get_nominal_value_for_product,
        EXPR.NPV_ProductExpression: _get_nominal_value_for_product,
        EXPR.DivisionExpression: _get_nominal_value_for_division,
        EXPR.NPV_DivisionExpression: _get_nominal_value_for_division,
        EXPR.PowExpression: _get_nominal_value_for_power,
        EXPR.NPV_PowExpression: _get_nominal_value_for_power,
        EXPR.NegationExpression: _get_nominal_value_negation,
        EXPR.NPV_NegationExpression: _get_nominal_value_negation,
        EXPR.AbsExpression: _get_nominal_value_abs,
        EXPR.NPV_AbsExpression: _get_nominal_value_abs,
        EXPR.UnaryFunctionExpression: _get_nominal_value_for_unary_function,
        EXPR.NPV_UnaryFunctionExpression: _get_nominal_value_for_unary_function,
        EXPR.Expr_ifExpression: _get_nominal_value_expr_if,
        EXPR.ExternalFunctionExpression: _get_nominal_value_external_function,
        EXPR.NPV_ExternalFunctionExpression: _get_nominal_value_external_function,
        EXPR.LinearExpression: _get_nominal_value_for_sum,
    }

    def exitNode(self, node, data):
        """Callback for :class:`pyomo.core.current.StreamBasedExpressionVisitor`. This
        method is called when moving back up the tree in a depth first search."""

        # first check if the node is a leaf
        nodetype = type(node)

        if nodetype in native_types:
            return [node]

        node_func = self.node_type_method_map.get(nodetype, None)
        if node_func is not None:
            return node_func(self, node, data)

        elif not node.is_expression_type():
            # this is a leaf, but not a native type
            if nodetype is _PyomoUnit:
                return [1]
            else:
                return self._get_magnitude_base_type(node)
                # might want to add other common types here

        # not a leaf - check if it is a named expression
        if (
            hasattr(node, "is_named_expression_type")
            and node.is_named_expression_type()
        ):
            return self._get_nominal_value_single_child(node, data)

        raise TypeError(
            f"An unhandled expression node type: {str(nodetype)} was encountered while "
            f"retrieving the nominal value of expression {str(node)}"
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for thickener unit model.
Authors: Andrew Lee
"""
from math import isnan
import pytest
import re

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Param,
    units,
    value,
    Var,
)
from pyomo.network import Port

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    MomentumBalanceType,
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    LiquidPhase,
    SolidPhase,
    Component,
)
from idaes.models.unit_models.solid_liquid import Thickener0D
from idaes.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox
from idaes.models.unit_models.separator import (
    SeparatorData,
    EnergySplittingType,
)
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


@declare_process_block_class("SolidParameterBlock")
class SolidParameterData(PhysicalParameterBlock):
    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._state_block_class = SolidStateBlock

        # Add Phase objects
        self.sol = SolidPhase()

        # Add Component objects
        self.a = Component()
        self.b = Component()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units(
            {
                "time": units.s,
                "length": units.m,
                "mass": units.kg,
                "amount": units.mol,
                "temperature": units.K,
            }
        )


@declare_process_block_class("SolidStateBlock", block_class=StateBlock)
class SolidStateBlockData(StateBlockData):
    def build(self):
        super().build()

        # Create state variables
        self.flow_mass = Var(
            initialize=1.0,
            units=units.kg / units.s,
        )
        self.flow_vol = Var(
            initialize=1.0,
            units=units.m**3 / units.s,
        )
        self.pressure = Var(
            initialize=101325.0,
            bounds=(1e3, 1e6),
            units=units.Pa,
        )
        self.temperature = Var(
            initialize=298.15,
            bounds=(298.15, 323.15),
            units=units.K,
        )
        self.mass_frac_comp = Var(
            self.params.component_list,
            initialize=0.1,
            units=units.dimensionless,
        )

        self.dens_mass = Param(
            initialize=2500,
            units=units.kg / units.m**3,
        )

        if self.config.defined_state is False:
            self.sum_mass_frac_eqn = Constraint(
                expr=sum(self.mass_frac_comp[j] for j in self.component_list) == 1
            )

        self.volumetric_flow = Constraint(
            expr=self.flow_vol * self.dens_mass == self.flow_mass
        )

    def get_material_flow_terms(b, p, j):
        return b.flow_mass * b.mass_frac_comp[j]

    def get_enthalpy_flow_terms(b, p):
        return (
            b.flow_mass * 42e3 * units.J / units.kg * (b.temperature - 273.15 * units.K)
        )

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def define_state_vars(b):
        return {
            "flow_mass": b.flow_mass,
            "mass_frac_comp": b.mass_frac_comp,
            "temperature": b.temperature,
            "pressure": b.pressure,
        }

    def get_material_flow_basis(b):
        return MaterialFlowBasis.mass


@declare_process_block_class("LiquidParameterBlock")
class LiquidParameterData(PhysicalParameterBlock):
    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._state_block_class = LiquidStateBlock

        # Add Phase objects
        self.liq = LiquidPhase()

        # Add Component objects
        self.c = Component()
        self.d = Component()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units(
            {
                "time": units.s,
                "length": units.m,
                "mass": units.kg,
                "amount": units.mol,
                "temperature": units.K,
            }
        )


@declare_process_block_class("LiquidStateBlock", block_class=StateBlock)
class LiquidStateBlockData(StateBlockData):
    def build(self):
        super().build()

        # Create state variables
        self.flow_mass = Var(
            initialize=1.0,
            units=units.kg / units.s,
        )
        self.flow_vol = Var(
            initialize=1.0,
            units=units.m**3 / units.s,
        )
        self.pressure = Var(
            initialize=101325.0,
            bounds=(1e3, 1e6),
            units=units.Pa,
        )
        self.temperature = Var(
            initialize=298.15,
            bounds=(298.15, 323.15),
            units=units.K,
        )
        self.mass_frac_comp = Var(
            self.params.component_list,
            initialize=0.1,
            units=units.dimensionless,
        )

        self.dens_mass = Param(
            initialize=1000,
            units=units.kg / units.m**3,
        )
        self.visc_d = Param(
            initialize=1e-3,
            units=units.Pa * units.s,
        )

        if self.config.defined_state is False:
            self.sum_mass_frac_eqn = Constraint(
                expr=sum(self.mass_frac_comp[j] for j in self.component_list) == 1
            )

        self.volumetric_flow = Constraint(
            expr=self.flow_vol * self.dens_mass == self.flow_mass
        )

    def get_material_flow_terms(b, p, j):
        return b.flow_mass * b.mass_frac_comp[j]

    def get_enthalpy_flow_terms(b, p):
        return (
            b.flow_mass * 42e3 * units.J / units.kg * (b.temperature - 273.15 * units.K)
        )

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def define_state_vars(b):
        return {
            "flow_mass": b.flow_mass,
            "mass_frac_comp": b.mass_frac_comp,
            "temperature": b.temperature,
            "pressure": b.pressure,
        }

    def get_material_flow_basis(b):
        return MaterialFlowBasis.mass


# -----------------------------------------------------------------------------
class TestThickener0DBasic:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.solid = SolidParameterBlock()
        m.fs.liquid = LiquidParameterBlock()

        m.fs.unit = Thickener0D(
            solid_property_package=m.fs.solid,
            liquid_property_package=m.fs.liquid,
        )

        m.fs.unit.solid_inlet.flow_mass.fix(0.7 * 7e-6 * 2500)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "a"].fix(0.2)
        m.fs.unit.solid_inlet.mass_frac_comp[0, "b"].fix(0.8)
        m.fs.unit.solid_inlet.temperature.fix(303.15)
        m.fs.unit.solid_inlet.pressure.fix(101325.0)

        m.fs.unit.liquid_inlet.flow_mass.fix(0.3 * 7e-6 * 1000)
        m.fs.unit.liquid_inlet.mass_frac_comp[0, "c"].fix(0.3)
        m.fs.unit.liquid_inlet.mass_frac_comp[0, "d"].fix(0.7)
        m.fs.unit.liquid_inlet.temperature.fix(310)
        m.fs.unit.liquid_inlet.pressure.fix(1.5e5)

        m.fs.unit.flow_vol_underflow.fix(5e-6)

        # Parameters
        m.fs.unit.area.fix(1)
        m.fs.unit.v0.fix(1.18e-4)
        m.fs.unit.v1.fix(1e-5)
        m.fs.unit.C.fix(5)
        m.fs.unit.solid_fraction_max.fix(0.9)

        return m

    @pytest.mark.unit
    def test_config(self, model):
        # Check unit config arguments
        assert len(model.fs.unit.config) == 9

        assert not model.fs.unit.config.dynamic
        assert not model.fs.unit.config.has_holdup
        assert (
            model.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
        )
        assert (
            model.fs.unit.config.energy_split_basis
            == EnergySplittingType.equal_temperature
        )
        assert (
            model.fs.unit.config.momentum_balance_type
            == MomentumBalanceType.pressureTotal
        )

        assert model.fs.unit.config.solid_property_package is model.fs.solid
        assert model.fs.unit.config.liquid_property_package is model.fs.liquid

        assert model.fs.unit.default_initializer is BlockTriangularizationInitializer

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.solid_inlet, Port)
        assert isinstance(model.fs.unit.solid_underflow, Port)
        assert isinstance(model.fs.unit.solid_overflow, Port)

        assert isinstance(model.fs.unit.liquid_inlet, Port)
        assert isinstance(model.fs.unit.liquid_underflow, Port)
        assert isinstance(model.fs.unit.liquid_overflow, Port)

        assert isinstance(model.fs.unit.solid_split, SeparatorData)
        assert isinstance(model.fs.unit.liquid_split, SeparatorData)

        assert number_variables(model) == 54
        assert number_total_constraints(model) == 40
        assert number_unused_variables(model) == 0

    @pytest.mark.component
    def test_structural_issues(self, model):
        dt = DiagnosticsToolbox(model)
        dt.assert_no_structural_warnings()

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=2e-5)
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, model):
        dt = DiagnosticsToolbox(model)
        dt.assert_no_numerical_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert value(model.fs.unit.solid_fraction_feed[0]) == pytest.approx(
            0.7, rel=1e-5
        )
        assert value(model.fs.unit.solid_fraction_underflow[0]) == pytest.approx(
            0.816954, rel=1e-5
        )
        assert value(model.fs.unit.solid_fraction_overflow[0]) == pytest.approx(
            0.407615, rel=1e-5
        )
        assert value(model.fs.unit.particle_size[0]) == pytest.approx(
            1.20163e-5, rel=1e-5
        )
        so = model.fs.unit.solid_split.overflow_state[0].flow_vol
        lo = model.fs.unit.liquid_split.overflow_state[0].flow_vol
        su = model.fs.unit.solid_split.underflow_state[0].flow_vol
        lu = model.fs.unit.liquid_split.underflow_state[0].flow_vol

        assert value(su / (lu + su)) == pytest.approx(
            value(model.fs.unit.solid_fraction_underflow[0]), rel=1e-5
        )
        assert value(so / (lo + so)) == pytest.approx(
            value(model.fs.unit.solid_fraction_overflow[0]), rel=1e-5
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        sf = model.fs.unit.solid_inlet_state[0]
        lf = model.fs.unit.liquid_inlet_state[0]
        so = model.fs.unit.solid_split.overflow_state[0]
        lo = model.fs.unit.liquid_split.overflow_state[0]
        su = model.fs.unit.solid_split.underflow_state[0]
        lu = model.fs.unit.liquid_split.underflow_state[0]

        # Solid mass conservation
        assert value(sf.flow_mass - so.flow_mass - su.flow_mass) <= 1e-8
        # Liquid mass conservation
        assert value(lf.flow_mass - lo.flow_mass - lu.flow_mass) <= 1e-8

        # Solid volume conservation
        assert value(sf.flow_vol - so.flow_vol - su.flow_vol) <= 1e-8
        # Liquid volume conservation
        assert value(lf.flow_vol - lo.flow_vol - lu.flow_vol) <= 1e-8

        assert value(so.temperature) == pytest.approx(303.15, rel=1e-5)
        assert value(su.temperature) == pytest.approx(303.15, rel=1e-5)
        assert value(so.pressure) == pytest.approx(101325, rel=1e-5)
        assert value(su.pressure) == pytest.approx(101325, rel=1e-5)

        assert value(lo.temperature) == pytest.approx(310, rel=1e-5)
        assert value(lu.temperature) == pytest.approx(310, rel=1e-5)
        assert value(lo.pressure) == pytest.approx(1.5e5, rel=1e-5)
        assert value(lu.pressure) == pytest.approx(1.5e5, rel=1e-5)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, model):
        perf_dict = model.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Area": model.fs.unit.area,
                "Liquid Recovery": model.fs.unit.liquid_split.split_fraction[
                    0, "overflow"
                ],
                "Feed Solid Fraction": model.fs.unit.solid_fraction_feed[0],
                "Underflow Solid Fraction": model.fs.unit.solid_fraction_underflow[0],
                "particle size": model.fs.unit.particle_size[0],
                "v0": model.fs.unit.v0[0],
                "v1": model.fs.unit.v1,
                "C": model.fs.unit.C,
                "solid_fraction_max": model.fs.unit.solid_fraction_max,
            },
        }

    @pytest.mark.ui
    @pytest.mark.component
    def test_get_stream_table_contents(self, model):
        stable = model.fs.unit._get_stream_table_contents()

        expected = {
            "Units": {
                "flow_mass": getattr(units.pint_registry, "kg/second"),
                "mass_frac_comp a": getattr(units.pint_registry, "dimensionless"),
                "mass_frac_comp b": getattr(units.pint_registry, "dimensionless"),
                "temperature": getattr(units.pint_registry, "K"),
                "pressure": getattr(units.pint_registry, "Pa"),
                "mass_frac_comp c": getattr(units.pint_registry, "dimensionless"),
                "mass_frac_comp d": getattr(units.pint_registry, "dimensionless"),
            },
            "Feed Solid": {
                "flow_mass": pytest.approx(0.01225, rel=1e-4),
                "mass_frac_comp a": pytest.approx(0.2, rel=1e-4),
                "mass_frac_comp b": pytest.approx(0.8, rel=1e-4),
                "temperature": pytest.approx(303.15, rel=1e-4),
                "pressure": pytest.approx(101325, rel=1e-4),
                "mass_frac_comp c": float("nan"),
                "mass_frac_comp d": float("nan"),
            },
            "Feed Liquid": {
                "flow_mass": pytest.approx(0.0021, rel=1e-4),
                "mass_frac_comp a": float("nan"),
                "mass_frac_comp b": float("nan"),
                "temperature": pytest.approx(310, rel=1e-4),
                "pressure": pytest.approx(1.5e5, rel=1e-4),
                "mass_frac_comp c": pytest.approx(0.3, rel=1e-4),
                "mass_frac_comp d": pytest.approx(0.7, rel=1e-4),
            },
            "Underflow Solid": {
                "flow_mass": pytest.approx(0.010212, rel=1e-4),
                "mass_frac_comp a": pytest.approx(0.2, rel=1e-4),
                "mass_frac_comp b": pytest.approx(0.8, rel=1e-4),
                "temperature": pytest.approx(303.15, rel=1e-4),
                "pressure": pytest.approx(101325, rel=1e-4),
                "mass_frac_comp c": float("nan"),
                "mass_frac_comp d": float("nan"),
            },
            "Underflow Liquid": {
                "flow_mass": pytest.approx(0.00091523, rel=1e-4),
                "mass_frac_comp a": float("nan"),
                "mass_frac_comp b": float("nan"),
                "temperature": pytest.approx(310, rel=1e-4),
                "pressure": pytest.approx(1.5e5, rel=1e-4),
                "mass_frac_comp c": pytest.approx(0.3, rel=1e-4),
                "mass_frac_comp d": pytest.approx(0.7, rel=1e-4),
            },
            "Overflow Solid": {
                "flow_mass": pytest.approx(0.0020381, rel=1e-4),
                "mass_frac_comp a": pytest.approx(0.2, rel=1e-4),
                "mass_frac_comp b": pytest.approx(0.8, rel=1e-4),
                "temperature": pytest.approx(303.15, rel=1e-4),
                "pressure": pytest.approx(101325, rel=1e-4),
                "mass_frac_comp c": float("nan"),
                "mass_frac_comp d": float("nan"),
            },
            "Overflow Liquid": {
                "flow_mass": pytest.approx(0.0011848, rel=1e-4),
                "mass_frac_comp a": float("nan"),
                "mass_frac_comp b": float("nan"),
                "temperature": pytest.approx(310, rel=1e-4),
                "pressure": pytest.approx(1.5e5, rel=1e-4),
                "mass_frac_comp c": pytest.approx(0.3, rel=1e-4),
                "mass_frac_comp d": pytest.approx(0.7, rel=1e-4),
            },
        }

        stb = stable.to_dict()
        for k, d in stb.items():
            if k == "Units":
                assert d == expected["Units"]
            else:
                for i, dd in d.items():
                    if not isnan(dd):
                        assert dd == expected[k][i]
                    else:
                        if "Liquid" in k:
                            assert i in ["mass_frac_comp a", "mass_frac_comp b"]
                        else:
                            assert i in ["mass_frac_comp c", "mass_frac_comp d"]

    @pytest.mark.unit
    def test_deprecate_initialize(self, model):
        with pytest.raises(
            NotImplementedError,
            match=re.escape(
                "The Thickener0D unit model does not support the old initialization API. "
                "Please use the new API (InitializerObjects) instead."
            ),
        ):
            model.fs.unit.initialize()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Interface for importing Keras models into IDAES
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import json
import os.path

import pandas as pd

from pyomo.common.dependencies import attempt_import

from idaes.core.surrogate.sampling.scaling import OffsetScaler

from idaes.core.surrogate.omlt_base_surrogate_class import OMLTSurrogate

keras, keras_available = attempt_import("tensorflow.keras")
omlt, omlt_available = attempt_import("omlt")

if omlt_available:
    from omlt.neuralnet.nn_formulation import (
        FullSpaceSmoothNNFormulation,
        ReducedSpaceSmoothNNFormulation,
        ReluBigMFormulation,
        ReluComplementarityFormulation,
    )

    if keras_available:
        from omlt.io import load_keras_sequential


class KerasSurrogate(OMLTSurrogate):
    def __init__(
        self,
        keras_model,
        input_labels,
        output_labels,
        input_bounds,
        input_scaler=None,
        output_scaler=None,
    ):
        """
        Standard SurrogateObject for surrogates based on Keras models.
        Utilizes the OMLT framework for importing Keras models to IDAES.

        Contains methods to both populate a Pyomo Block with constraints
        representing the surrogate and to evaluate the surrogate a set of user
        provided points.

        This constructor should only be used when first creating the surrogate within IDAES.
        Once created, this object can be stored to disk using save_to_folder and loaded
        with load_from_folder

        Args:
           keras_model: Keras Sequential model
              This is the Keras Sequential model that will be loaded. Note that
              specialized layers may not be supported at this time.
           input_labels: list of str
              The ordered list of labels corresponding to the inputs in the keras model
           output_labels: list of str
              The ordered list of labels corresponding to the outputs in the keras model
           input_bounds: None of dict of tuples
              Keys correspond to each of the input labels and values are the tuples of
              bounds (lb, ub)
           input_scaler: None or OffsetScaler
              The scaler to be used for the inputs. If None, then no scaler is used
           output_scaler: None of OffsetScaler
              The scaler to be used for the outputs. If None, then no scaler is used
        """
        super().__init__(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=input_bounds,
            input_scaler=input_scaler,
            output_scaler=output_scaler,
        )
        self._keras_model = keras_model

    def populate_block(self, block, additional_options=None):
        """
        Method to populate a Pyomo Block with the keras model constraints.

        Args:
           block: Pyomo Block component
              The block to be populated with variables and/or constraints.
           additional_options: dict or None
              If not None, then should be a dict with the following keys;
              'formulation': KerasSurrogate.Formulation
              The formulation to use with OMLT. Possible values are FULL_SPACE,
              REDUCED_SPACE, RELU_BIGM, or RELU_COMPLEMENTARITY (default is FULL_SPACE)
        """
        formulation = additional_options.pop(
            "formulation", KerasSurrogate.Formulation.FULL_SPACE
        )
        omlt_scaling, scaled_input_bounds = self.generate_omlt_scaling_objecets()

        net = load_keras_sequential(
            self._keras_model,
            scaling_object=omlt_scaling,
            scaled_input_bounds=scaled_input_bounds,
        )

        if formulation == KerasSurrogate.Formulation.FULL_SPACE:
            formulation_object = FullSpaceSmoothNNFormulation(net)
        elif formulation == KerasSurrogate.Formulation.REDUCED_SPACE:
            formulation_object = ReducedSpaceSmoothNNFormulation(net)
        elif formulation == KerasSurrogate.Formulation.RELU_BIGM:
            formulation_object = ReluBigMFormulation(net)
        elif formulation == KerasSurrogate.Formulation.RELU_COMPLEMENTARITY:
            formulation_object = ReluComplementarityFormulation(net)
        else:
            raise ValueError(
                'An unrecognized formulation "{}" was passed to '
                "KerasSurrogate.populate_block. Please pass a valid "
                "formulation.".format(formulation)
            )
        self.populate_block_with_net(block, formulation_object)

    def evaluate_surrogate(self, inputs):
        """
        Method to evaluate Keras model at a set of input values.

        Args:
            inputs: numpy array of input values. First dimension of array
                must match the number of input variables.

        Returns:
            outputs: numpy array of values for all outputs evaluated at input
                points.
        """
        x = inputs
        if self._input_scaler is not None:
            x = self._input_scaler.scale(x)
        y = self._keras_model.predict(x.to_numpy())

        # y is a numpy array, make it a dataframe
        y = pd.DataFrame(
            data=y, columns=self.output_labels(), index=inputs.index, dtype="float64"
        )
        if self._output_scaler is not None:
            y = self._output_scaler.unscale(y)
        return y

    def save_to_folder(self, keras_folder_name, keras_model_name="idaes_keras_model"):
        """
        Save the surrogate object to disk by providing the name of the
        folder to contain the keras model and additional IDAES metadata

        Args:
           folder_name: str
              The name of the folder to contain the Keras model and additional
              IDAES metadata
        """
        self._keras_model.save(
            os.path.join(keras_folder_name, keras_model_name + ".keras")
        )
        info = dict()
        info["input_scaler"] = None
        if self._input_scaler is not None:
            info["input_scaler"] = self._input_scaler.to_dict()
        info["output_scaler"] = None
        if self._output_scaler is not None:
            info["output_scaler"] = self._output_scaler.to_dict()

        # serialize information from the base class
        info["input_labels"] = self.input_labels()
        info["output_labels"] = self.output_labels()
        info["input_bounds"] = self.input_bounds()

        with open(os.path.join(keras_folder_name, "idaes_info.json"), "w") as fd:
            json.dump(info, fd)

    @classmethod
    def load_from_folder(cls, keras_folder_name, keras_model_name="idaes_keras_model"):
        """
        Load the surrogate object from disk by providing the name of the
        folder holding the keras model

        Args:
           folder_name: str
              The name of the folder containing the Keras model and additional
              IDAES metadata

        Returns: an instance of KerasSurrogate
        """

        keras_model = keras.models.load_model(
            os.path.join(keras_folder_name, keras_model_name + ".keras")
        )

        with open(os.path.join(keras_folder_name, "idaes_info.json")) as fd:
            info = json.load(fd)

        input_scaler = None
        if info["input_scaler"] is not None:
            input_scaler = OffsetScaler.from_dict(info["input_scaler"])

        output_scaler = None
        if info["output_scaler"] is not None:
            output_scaler = OffsetScaler.from_dict(info["output_scaler"])

        return KerasSurrogate(
            keras_model=keras_model,
            input_labels=info["input_labels"],
            output_labels=info["output_labels"],
            input_bounds=info["input_bounds"],
            input_scaler=input_scaler,
            output_scaler=output_scaler,
        )


def save_keras_json_hd5(nn, path, name):
    json_model = nn.to_json()
    with open(os.path.join(path, "{}.json".format(name)), "w") as json_file:
        json_file.write(json_model)
    nn.save(os.path.join(path, "{}.keras".format(name)))
    nn.save_weights(os.path.join(path, "{}.weights.h5".format(name)))


def load_keras_json_hd5(path, name):
    nn = keras.models.load_model(os.path.join(path, "{}.keras".format(name)))
    nn.load_weights(os.path.join(path, "{}.weights.h5".format(name)))
    return nn



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee
"""

import pytest

from numpy import logspace

from pyomo.util.check_units import assert_units_consistent
from pyomo.environ import assert_optimal_termination, ConcreteModel, Objective, value

from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available
from idaes.models.properties.modular_properties.examples.BT_PR import configuration
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.models.properties.tests.test_harness import PropertyTestHarness

import idaes.logger as idaeslog

SOUT = idaeslog.INFO

# Set module level pyest marker
pytestmark = pytest.mark.cubic_root


# -----------------------------------------------------------------------------
# Get default solver for testing
# Limit iterations to make sure sweeps aren't getting out of hand
solver = get_solver(solver="ipopt_v2", solver_options={"max_iter": 100})


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBTPR(PropertyTestHarness):
    def configure(self):
        self.prop_pack = GenericParameterBlock
        self.param_args = configuration
        self.prop_args = {}
        self.has_density_terms = False


# -----------------------------------------------------------------------------
# Test robustness and some outputs
@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBTExample(object):
    @pytest.fixture()
    def m(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = GenericParameterBlock(**configuration)

        m.fs.state = m.fs.props.build_state_block([1], defined_state=True)

        # Set small values of epsilon to get accurate results
        # Initialization will handle finding the correct region
        m.fs.state[1].eps_t_Vap_Liq.set_value(1e-4)
        m.fs.state[1].eps_z_Vap_Liq.set_value(1e-4)

        iscale.calculate_scaling_factors(m.fs.props)
        iscale.calculate_scaling_factors(m.fs.state[1])
        return m

    @pytest.mark.integration
    def test_T_sweep(self, m):
        assert_units_consistent(m)

        m.fs.obj = Objective(expr=(m.fs.state[1].temperature - 510) ** 2)
        m.fs.state[1].temperature.setub(600)

        for P in logspace(4.8, 5.9, 8):

            m.fs.state[1].flow_mol.fix(100)
            m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
            m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
            m.fs.state[1].temperature.fix(300)
            m.fs.state[1].pressure.fix(P)

            # For optimization sweep, use a large eps to avoid getting stuck at
            # bubble and dew points
            m.fs.state[1].eps_t_Vap_Liq.set_value(10)
            m.fs.state[1].eps_z_Vap_Liq.set_value(10)

            m.fs.state.initialize()

            m.fs.state[1].temperature.unfix()
            m.fs.obj.activate()

            results = solver.solve(m)
            assert_optimal_termination(results)

            # Switch to small eps and re-solve to refine result
            m.fs.state[1].eps_t_Vap_Liq.set_value(1e-4)
            m.fs.state[1].eps_z_Vap_Liq.set_value(1e-4)

            results = solver.solve(m)

            assert_optimal_termination(results)
            assert m.fs.state[1].flow_mol_phase["Liq"].value <= 1e-5

    @pytest.mark.integration
    def test_P_sweep(self, m):
        for T in range(370, 500, 25):
            m.fs.state[1].flow_mol.fix(100)
            m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
            m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
            m.fs.state[1].temperature.fix(T)
            m.fs.state[1].pressure.fix(1e5)

            m.fs.state.initialize()

            results = solver.solve(m)

            assert_optimal_termination(results)

            while m.fs.state[1].pressure.value <= 1e6:

                results = solver.solve(m)
                assert_optimal_termination(results)

                m.fs.state[1].pressure.value = m.fs.state[1].pressure.value + 1e5

    @pytest.mark.component
    def test_T350_P1_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(350)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), abs=1e-1) == 365
        assert 0.0035346 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.966749 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.894676
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.347566
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.971072
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.959791
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.70584
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.29416
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 38942.8
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 78048.7
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -361.794
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -264.0181
        )

    @pytest.mark.component
    def test_T350_P5_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(350)
        m.fs.state[1].pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 431.47
        assert (
            pytest.approx(value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5)
            == 0.01766
        )
        assert (
            pytest.approx(value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5)
            == 0.80245
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.181229
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.070601
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.856523
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.799237
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.65415
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.34585
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 38966.9
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 75150.7
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -361.8433
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -281.9703
        )

    @pytest.mark.component
    def test_T450_P1_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(450)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 371.4
        assert 0.0033583 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.9821368 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 8.069323
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 4.304955
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.985365
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.979457
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.29861
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.70139
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.5
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 49441.2
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 84175.1
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -328.766
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -241.622
        )

    @pytest.mark.component
    def test_T450_P5_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(450)
        m.fs.state[1].pressure.fix(5e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 436.93
        assert 0.0166181 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.9053766 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.63308
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.873213
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.927534
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.898324
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.3488737
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.6511263
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.5
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 51095.2
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 83362.3
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -326.299
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -256.198
        )

    @pytest.mark.component
    def test_T368_P1_x5(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.5)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.5)
        m.fs.state[1].temperature.fix(368)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 368
        assert 0.003504 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.97 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.492049
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.621563
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.97469
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.964642
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.4012128
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.5987872
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.6141738
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.3858262
        )

        m.fs.state[1].mole_frac_phase_comp.display()
        m.fs.state[1].enth_mol_phase_comp.display()

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 38235.1
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 77155.4
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -359.256
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -262.348
        )

    @pytest.mark.component
    def test_T376_P1_x2(self, m):
        m.fs.state[1].flow_mol.fix(100)
        m.fs.state[1].mole_frac_comp["benzene"].fix(0.2)
        m.fs.state[1].mole_frac_comp["toluene"].fix(0.8)
        m.fs.state[1].temperature.fix(376)
        m.fs.state[1].pressure.fix(1e5)

        # Trigger build of enthalpy and entropy
        m.fs.state[1].enth_mol_phase
        m.fs.state[1].entr_mol_phase

        m.fs.state.initialize(outlvl=SOUT)

        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

        assert pytest.approx(value(m.fs.state[1]._teq[("Vap", "Liq")]), 1e-5) == 376
        assert 0.00361333 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Liq"]), 1e-5
        )
        assert 0.968749 == pytest.approx(
            value(m.fs.state[1].compress_fact_phase["Vap"]), 1e-5
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 1.8394188
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.7871415
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.9763608
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].fug_coeff_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.9663611
        )

        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "benzene"]), 1e-5
            )
            == 0.17342
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Liq", "toluene"]), 1e-5
            )
            == 0.82658
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "benzene"]), 1e-5
            )
            == 0.3267155
        )
        assert (
            pytest.approx(
                value(m.fs.state[1].mole_frac_phase_comp["Vap", "toluene"]), 1e-5
            )
            == 0.6732845
        )

        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Liq"]), 1e-5) == 31535.8
        )
        assert (
            pytest.approx(value(m.fs.state[1].enth_mol_phase["Vap"]), 1e-5) == 69175.3
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Liq"]), 1e-5) == -369.033
        )
        assert (
            pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1e-5) == -273.513
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Steam turbine outlet stage model.  This model is based on:

Liese, (2014). "Modeling of a Steam Turbine Including Partial Arc Admission
    for Use in a Process Simulation Software Environment." Journal of Engineering
    for Gas Turbines and Power. v136.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

__Author__ = "John Eslick"

from pyomo.environ import Var, sqrt, value, units as pyunits
from idaes.models_extra.power_generation.unit_models.helm.turbine import (
    HelmIsentropicTurbineData,
)
from idaes.core import declare_process_block_class
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale

import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


@declare_process_block_class(
    "HelmTurbineOutletStage",
    doc="Outlet stage steam turbine model",
)
class HelmTurbineOutletStageData(HelmIsentropicTurbineData):
    # Same settings as the default pressure changer, but force to expander with
    # isentropic efficiency
    CONFIG = HelmIsentropicTurbineData.CONFIG()

    def build(self):
        super().build()

        self.flow_coeff = Var(
            initialize=0.0333,
            doc="Turbine flow coefficient [kg*C^0.5/s/Pa]",
            units=pyunits.kg * pyunits.K**0.5 / pyunits.s / pyunits.Pa,
        )
        self.eff_dry = Var(initialize=0.87, doc="Turbine dry isentropic efficiency")
        self.design_exhaust_flow_vol = Var(
            initialize=6000.0,
            doc="Design exit volumetirc flowrate [m^3/s]",
            units=pyunits.m**3 / pyunits.s,
        )
        self.efficiency_mech = Var(initialize=1.0, doc="Turbine mechanical efficiency")
        self.efficiency_isentropic.unfix()
        self.eff_dry.fix()
        self.design_exhaust_flow_vol.fix()
        self.flow_coeff.fix()
        self.efficiency_mech.fix()

        self.tel_c0 = Var(
            initialize=0.0064 * 1e6,
            units=pyunits.J / pyunits.mol,
            doc="c0 in tel = c0 + c1*fr + c2*fr**2 + ... + c5*fr**5 (fr is ratio"
            " of exhaust volumetric flow to design exhaust volumetric flow)",
        )
        self.tel_c1 = Var(
            initialize=-0.0328 * 1e6,
            units=pyunits.J / pyunits.mol,
            doc="c1 in tel = c0 + c1*fr + c2*fr**2 + ... + c5*fr**5 (fr is ratio"
            " of exhaust volumetric flow to design exhaust volumetric flow)",
        )
        self.tel_c2 = Var(
            initialize=0.0638 * 1e6,
            units=pyunits.J / pyunits.mol,
            doc="c2 in tel = c0 + c1*fr + c2*fr**2 + ... + c5*fr**5 (fr is ratio"
            " of exhaust volumetric flow to design exhaust volumetric flow)",
        )
        self.tel_c3 = Var(
            initialize=-0.0542 * 1e6,
            units=pyunits.J / pyunits.mol,
            doc="c3 in tel = c0 + c1*fr + c2*fr**2 + ... + c5*fr**5 (fr is ratio"
            " of exhaust volumetric flow to design exhaust volumetric flow)",
        )
        self.tel_c4 = Var(
            initialize=0.022 * 1e6,
            units=pyunits.J / pyunits.mol,
            doc="c4 in tel = c0 + c1*fr + c2*fr**2 + ... + c5*fr**5 (fr is ratio"
            " of exhaust volumetric flow to design exhaust volumetric flow)",
        )
        self.tel_c5 = Var(
            initialize=-0.0035 * 1e6,
            units=pyunits.J / pyunits.mol,
            doc="c5 in tel = c0 + c1*fr + c2*fr**2 + ... + c5*fr**5 (fr is ratio"
            " of exhaust volumetric flow to design exhaust volumetric flow)",
        )
        self.tel_c0.fix()
        self.tel_c1.fix()
        self.tel_c2.fix()
        self.tel_c3.fix()
        self.tel_c4.fix()
        self.tel_c5.fix()

        @self.Expression(self.flowsheet().time, doc="Total exhaust loss curve")
        def tel(b, t):
            f = b.control_volume.properties_out[t].flow_vol / b.design_exhaust_flow_vol
            return (
                +self.tel_c5 * f**5
                + self.tel_c4 * f**4
                + self.tel_c3 * f**3
                + self.tel_c2 * f**2
                + self.tel_c1 * f
                + self.tel_c0
            )

        @self.Constraint(self.flowsheet().time, doc="Stodola eq. choked flow")
        def stodola_equation(b, t):
            flow = b.control_volume.properties_in[t].flow_mol
            mw = b.control_volume.properties_in[t].mw
            Tin = b.control_volume.properties_in[t].temperature
            Pin = b.control_volume.properties_in[t].pressure
            Pr = b.ratioP[t]
            cf = b.flow_coeff

            return flow**2 * mw**2 * (Tin) == (cf**2 * Pin**2 * (1 - Pr**2))

        @self.Constraint(self.flowsheet().time, doc="Efficiency correlation")
        def efficiency_correlation(b, t):
            x = b.control_volume.properties_out[t].vapor_frac
            eff = b.efficiency_isentropic[t]
            dh_isen = b.delta_enth_isentropic[t]
            tel = b.tel[t]
            return eff == b.eff_dry * x * (1 - 0.65 * (1 - x)) * (1 + tel / dh_isen)

        @self.Expression(self.flowsheet().time, doc="Thermodynamic power [J/s]")
        def power_thermo(b, t):
            return b.control_volume.work[t]

        @self.Expression(self.flowsheet().time, doc="Shaft power [J/s]")
        def power_shaft(b, t):
            return b.power_thermo[t] * b.efficiency_mech

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        calculate_cf=True,
    ):
        """
        Initialize the outlet turbine stage model.  This deactivates the
        specialized constraints, then does the isentropic turbine initialization,
        then reactivates the constraints and solves.

        Args:
            outlvl : sets output level of initialization routine
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        # sp is what to save to make sure state after init is same as the start
        #   saves value, fixed, and active state, doesn't load originally free
        #   values, this makes sure original problem spec is same but initializes
        #   the values of free vars
        for t in self.flowsheet().time:
            if self.outlet.pressure[t].fixed:
                self.ratioP[t] = value(self.outlet.pressure[t] / self.inlet.pressure[t])
                self.deltaP[t] = value(self.outlet.pressure[t] - self.inlet.pressure[t])

        # Deactivate special constraints
        self.stodola_equation.deactivate()
        self.efficiency_correlation.deactivate()
        self.efficiency_isentropic.fix()
        self.deltaP.unfix()
        self.ratioP.unfix()
        self.inlet.fix()
        self.outlet.unfix()

        super().initialize_build(outlvl=outlvl, solver=solver, optarg=optarg)

        for t in self.flowsheet().time:
            mw = self.control_volume.properties_in[t].mw
            Tin = self.control_volume.properties_in[t].temperature
            Pin = self.control_volume.properties_in[t].pressure
            Pr = self.ratioP[t]
            if not calculate_cf:
                cf = self.flow_coeff
                self.inlet.flow_mol[t].fix(
                    value(cf * Pin * sqrt(1 - Pr**2) / mw / sqrt(Tin))
                )

        super().initialize_build(outlvl=outlvl, solver=solver, optarg=optarg)
        self.control_volume.properties_out[:].pressure.fix()

        # Free eff_isen and activate special constraints
        self.efficiency_isentropic.unfix()
        self.outlet.pressure.fix()
        if calculate_cf:
            self.flow_coeff.unfix()
            self.inlet.flow_mol.unfix()
            self.inlet.flow_mol[0].fix()
            flow = self.control_volume.properties_in[0].flow_mol
            mw = self.control_volume.properties_in[0].mw
            Tin = self.control_volume.properties_in[0].temperature
            Pin = self.control_volume.properties_in[0].pressure
            Pr = self.ratioP[0]
            self.flow_coeff.value = value(flow * mw * sqrt(Tin / (1 - Pr**2)) / Pin)

        else:
            self.inlet.flow_mol.unfix()

        self.stodola_equation.activate()
        self.efficiency_correlation.activate()

        # Create solver
        slvr = get_solver(solver, optarg)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = slvr.solve(self, tee=slc.tee)
        init_log.info(
            "Initialization Complete (Outlet Stage): {}".format(idaeslog.condition(res))
        )

        # reload original spec
        if calculate_cf:
            cf = value(self.flow_coeff)
        from_json(self, sd=istate, wts=sp)
        if calculate_cf:
            # cf was probably fixed, so will have to set the value again here
            # if you ask for it to be calculated.
            self.flow_coeff = cf

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        for t, c in self.stodola_equation.items():
            s = (
                iscale.get_scaling_factor(
                    self.control_volume.properties_in[t].flow_mol,
                    default=1,
                    warning=True,
                )
                ** 2
            )
            iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from idaes.logger import getModelLogger

logging = getModelLogger("MatOptModel")

from pyomo.environ import *
from pyomo.core.base.var import VarData
from pyomo.core.expr.numeric_expr import (
    MonomialTermExpression,
    SumExpression,
    NegationExpression,
)
from ..util.util import isZero, areEqual

DBL_TOL = 1e-5
DEFAULT_BIG_M = 9999
DEFAULT_EPS = DBL_TOL


# ================================================
# ==========      UTILITY FUNCTIONS     ==========
# ================================================
def getLB(e):
    """Calculates the lower bound (if possible) of a linear expression.

    Args:
        e: A Pyomo expression, Pyomo variable, float, or int.

    Returns:
        (float) lower bound from interval arithmetic. None if negative infinity.

    """
    # Future work: use Pyomo function to achieve this functionality
    # return compute_bounds_on_expr(e)[0]
    if isinstance(e, VarData):
        if e.is_fixed():
            return value(e)
        else:
            return e.lb
    elif isinstance(e, MonomialTermExpression):
        assert isinstance(e.args[1], pyomo.core.base.var.VarData), (
            "This code relies on the assumption that the only variable "
            "in a monomial expression is the second argument"
        )
        if e.args[0] > 0:
            sub_expr_result = getLB(e.args[1])
            return e.args[0] * sub_expr_result if sub_expr_result is not None else None
        else:
            sub_expr_result = getUB(e.args[1])
            return e.args[0] * sub_expr_result if sub_expr_result is not None else None
    elif isinstance(e, SumExpression):
        result = 0.0
        for elem in e.args:
            elem_lb = getLB(elem)
            if elem_lb is not None:
                result += elem_lb
            else:
                return None
        return result
    elif isinstance(e, NegationExpression):
        elem_ub = getUB(e.args[0])
        return -elem_ub if elem_ub is not None else None
    elif isinstance(e, (float, int)):
        return e
    else:
        print(type(e))
        raise NotImplementedError(
            "Unsupported expression type for lower bound calculation. Please check getLB() function in "
            "pyomo_modeling.py for supported types."
        )


def getUB(e):
    """Calculates the upper bound (if possible) of a linear expression.

    Args:
        e: A Pyomo expression, Pyomo variable, float, or int.

    Returns:
        (float) upper bound from interval arithmetic. None if infinity.

    """
    # Future work: use Pyomo function to achieve this functionality
    # return compute_bounds_on_expr(e)[1]
    if isinstance(e, VarData):
        if e.is_fixed():
            return value(e)
        else:
            return e.ub
    elif isinstance(e, MonomialTermExpression):
        assert isinstance(e.args[1], pyomo.core.base.var.VarData), (
            "This code relies on the assumption that the only variable "
            "in a monomial expression is the second argument"
        )
        if e.args[0] > 0:
            sub_expr_result = getUB(e.args[1])
            return e.args[0] * sub_expr_result if sub_expr_result is not None else None
        else:
            sub_expr_result = getLB(e.args[1])
            return e.args[0] * sub_expr_result if sub_expr_result is not None else None
    elif isinstance(e, SumExpression):
        result = 0.0
        for elem in e.args:
            elem_ub = getUB(elem)
            if elem_ub is not None:
                result += elem_ub
            else:
                return None
        return result
    elif isinstance(e, NegationExpression):
        elem_lb = getLB(e.args[0])
        return -elem_lb if elem_lb is not None else None
    elif isinstance(e, (float, int)):
        return e
    else:
        print(type(e))
        raise NotImplementedError(
            "Unsupported expression type for upper bound calculation. Please check getUB() function in "
            "pyomo_modeling.py for supported types."
        )


# ================================================
# ==========   GENERIC MODEL ELEMENTS   ==========
# ================================================
def makeMyPyomoBaseModel(C, Atoms=None, Confs=None):
    """
    Make the Pyomo model for a basic materials design problem.

    Creates the basic sets and variables that make up the problem.
    All variables are created sparsely, so they are initialized
    only after they are actually referenced in a constraint. In
    this way, we smartly eliminate unnecessary variables and
    constraints.

    Basic Variables:
        Yi: Presence of building block at site i
        Xij: Presence of building blocks at both sites i and j
        Ci: Count of building block bonds to neighbors to site i

    Common Descriptors:
        Zi: Presence of target site at i

    Atom-Specific Variables:
        Yik: Presence of building block of type k at site i
        Xijkl: Presence of type k and l at sites i and j, respectively
        Cikl: Count of neighbors of type l next to site i with type k

    Conformation-Specific Descriptors:
        Zic: Presence of conformation c at site i

    The basic variables and atom-specific variables above are
    automatically encoded by calling addConsForGeneralVars.
    The descriptor variables Zi and Zic must be explicitly constrained.
    Some standard approaches are formalized in:

    * addConsBoundDescriptorsWithImpl
    * addConsIndFromDescriptors
    * addConsZicFromYi
    * addConsZicFromYiLifted
    * addConsZicFromYik
    * addConsZicFromYikLifted

    Additional variables and constraints  can be created and managed on
    a model-specific level.

    Args:
        C (Canvas): The design space over which to model the problem.
        Atoms (list<Atom/Any>): Optional, the set of building blocks.
            If present, generates a meaningful set for K building blocks.
            Else, only the general variables for presence/absence are
            meaningful. (Default value = None)
        Confs (list<Design>): Optional, the set of conformations to
            potentially use for indicator variables.
            (Default value = None)

    Returns:
        (ConcreteModel): Pyomo model with basic sets and variables initialized.

    """
    m = ConcreteModel()
    # Adding Model formulation information
    m.Canvas = C
    m.Ni = C.NeighborhoodIndexes
    m.Atoms = Atoms
    m.Confs = Confs
    # Adding Basic Sets
    m.nI = len(C)
    m.I = Set(initialize=range(m.nI), ordered=True)
    # Adding Basic Variables
    m.Yi = Var(m.I, domain=Binary, dense=False)
    m.Xij = Var(m.I, m.I, domain=Binary, dense=False)

    def _ruleCiBounds(m, i):
        return 0, len(m.Ni[i])

    m.Ci = Var(m.I, domain=NonNegativeIntegers, bounds=_ruleCiBounds, dense=False)
    # Adding Common Descriptors
    m.Zi = Var(m.I, domain=Binary, dense=False)
    # Adding Atoms Set
    m.nK = len(Atoms) if Atoms is not None else 0
    m.K = Set(initialize=Atoms)
    m.Yik = Var(m.I, m.K, domain=Binary, dense=False)
    m.Xijkl = Var(m.I, m.I, m.K, m.K, domain=Binary, dense=False)

    def _ruleCiklBounds(m, i, k, l):
        return 0, len(m.Ni[i])

    m.Cikl = Var(
        m.I, m.K, m.K, domain=NonNegativeIntegers, bounds=_ruleCiklBounds, dense=False
    )
    # Adding Confs Set
    m.nC = len(Confs) if Confs is not None else 0
    m.C = Set(initialize=range(m.nC))
    m.Zic = Var(m.I, m.C, domain=Binary, dense=False)
    return m


def _addConsCiFromCikl(m):
    m.AssignCiFromCikl = Constraint(m.I)
    for i in m.Ci.keys():
        m.AssignCiFromCikl.add(
            index=i,
            expr=(
                m.Ci[i]
                == sum(
                    m.Cikl[i, k, l]
                    for k in m.K
                    for l in m.K
                    if k is not None and l is not None
                )
            ),
        )


def _addConsCiFromXij(m):
    m.AssignCiFromXij = Constraint(m.I)
    for i in m.Ci.keys():
        m.AssignCiFromXij.add(
            index=i,
            expr=(m.Ci[i] == sum(m.Xij[i, j] for j in m.Ni[i] if j is not None)),
        )


def _addConsCiklFromXijkl(m):
    m.AssignCiklFromXijkl = Constraint(m.I, m.K, m.K)
    for i, k, l in m.Cikl.keys():
        m.AssignCiklFromXijkl.add(
            index=(i, k, l),
            expr=(
                m.Cikl[i, k, l]
                == sum(m.Xijkl[i, j, k, l] for j in m.Ni[i] if j is not None)
            ),
        )


def _addConsXijFromXijkl(m):
    m.AssignXijFromXijkl = Constraint(m.I, m.I)
    for i, j in m.Xij.keys():
        m.AssignXijFromXijkl.add(
            index=(i, j),
            expr=(
                m.Xij[i, j]
                == sum(
                    m.Xijkl[i, j, k, l]
                    for k in m.K
                    for l in m.K
                    if k is not None and l is not None
                )
            ),
        )


def _addConsXijFromYi(m):
    m.AssignXijFromYi1 = Constraint(m.I, m.I)
    m.AssignXijFromYi2 = Constraint(m.I, m.I)
    m.AssignXijFromYi3 = Constraint(m.I, m.I)
    for i, j in m.Xij.keys():
        m.AssignXijFromYi1.add(index=(i, j), expr=(m.Xij[i, j] <= m.Yi[i]))
        m.AssignXijFromYi2.add(index=(i, j), expr=(m.Xij[i, j] <= m.Yi[j]))
        m.AssignXijFromYi3.add(
            index=(i, j), expr=(m.Xij[i, j] >= m.Yi[i] + m.Yi[j] - 1)
        )


def _addConsXijklFromYik(m):
    m.AssignXijklFromYik1 = Constraint(m.I, m.I, m.K, m.K)
    m.AssignXijklFromYik2 = Constraint(m.I, m.I, m.K, m.K)
    m.AssignXijklFromYik3 = Constraint(m.I, m.I, m.K, m.K)
    for i, j, k, l in m.Xijkl.keys():
        m.AssignXijklFromYik1.add(
            index=(i, j, k, l), expr=(m.Xijkl[i, j, k, l] <= m.Yik[i, k])
        )
        m.AssignXijklFromYik2.add(
            index=(i, j, k, l), expr=(m.Xijkl[i, j, k, l] <= m.Yik[j, l])
        )
        m.AssignXijklFromYik3.add(
            index=(i, j, k, l),
            expr=(m.Xijkl[i, j, k, l] >= m.Yik[i, k] + m.Yik[j, l] - 1),
        )


def _addConsYiFromYik(m):
    m.AssignYiFromYik = Constraint(m.I)
    for i in m.Yi.keys():
        m.AssignYiFromYik.add(
            index=i, expr=(m.Yi[i] == sum(m.Yik[i, k] for k in m.K if k is not None))
        )


def _addConsYikSOS1(m):
    m.AssignYikSOS1 = Constraint(m.I)
    for i in m.Yik.index_set().set_tuple[0]:
        m.AssignYikSOS1.add(index=i, expr=(sum(m.Yik[i, k] for k in m.K) <= 1))


def addConsForGeneralVars(m):
    """Scan over the model and encode constraints for all basic variables present.

    Encodes variables in a chain from derived to most basc:
        Ci   <- Xij   <- Yi
        Cikl <- Xijkl <- Yik

    Also can bridge from type-specific to type-agnostic variables:
        Ci  <- Cikl
        Xij <- Xijkl
        Yi  <- Yik

    Args:
        m (ConcreteModel): Model to encode basic constraints.

    Returns:
        None.

    Raises:
        NotImplementedError: Several competing encodings exist and make
            the set of constraints not clearly defined.

    """
    # Defining Ci
    if len(m.Ci) > 0:
        if len(m.Xij) > 0 and len(m.Cikl) > 0:
            _addConsCiFromCikl(m)
            # NOTE: Chose to add only one formulation because the second is redundant
        elif len(m.Cikl) > 0 or len(m.Xijkl) > 0 or len(m.Yik) > 0:
            _addConsCiFromCikl(m)
        elif len(m.Xij) > 0 or len(m.Yi) > 0 or len(m.Yik) == 0:
            _addConsCiFromXij(m)
        else:
            raise NotImplementedError(
                "Ci lacks a proper way to be defined. User should define MaterialDescriptor Ci explicitly using valid "
                "DescriptorRule"
            )
            # Defining Cikl
    if len(m.Cikl) > 0:
        _addConsCiklFromXijkl(m)
    # Defining Xij
    if len(m.Xij) > 0:
        if len(m.Xijkl) > 0 and len(m.Yi) > 0:
            # NOTE: Chose to add both formulation because we believe neither is dominating
            _addConsXijFromXijkl(m)
            _addConsXijFromYi(m)
        elif len(m.Xijkl) > 0 or len(m.Yik) > 0:
            _addConsXijFromXijkl(m)
        elif len(m.Yi) > 0 or len(m.Yik) == 0:
            _addConsXijFromYi(m)
        else:
            raise NotImplementedError(
                "Xij lacks a proper way to be defined. User should define MaterialDescriptor Xij explicitly using "
                "valid DescriptorRule"
            )
            # Defining Xijkl
    if len(m.Xijkl) > 0:
        _addConsXijklFromYik(m)
    # Defining Zic
    if len(m.Zic) > 0:
        if m.nK > 1:
            addConsZicFromYikLifted(m)
        else:
            addConsZicFromYiLifted(m)
    # Defining Yi
    if len(m.Yi) > 0:
        if len(m.Yik) > 0:
            _addConsYiFromYik(m)
        else:
            # Yi can serve as a suitable basic variable if no 'k' index is in the model
            pass
    # Define Yik
    # Still need to restrict Yik to be SOS1
    if len(m.K) > 1:
        _addConsYikSOS1(m)


def fixYik(m, i, k, val):
    if isZero(val, DBL_TOL):
        fixYikDown(m, i, k)
    else:
        assert areEqual(val, 1.0, DBL_TOL)
        fixYikUp(m, i, k)


def fixYi(m, i, val):
    if isZero(val, DBL_TOL):
        fixYiDown(m, i)
    else:
        assert areEqual(val, 1.0, DBL_TOL)
        fixYiUp(m, i)


def fixXijkl(m, i, j, k, l, val):
    if isZero(val, DBL_TOL):
        fixXijklDown(m, i, j, k, l)
    else:
        assert areEqual(val, 1.0, DBL_TOL)
        fixXijklUp(m, i, j, k, l)


def fixXij(m, i, j, val):
    if isZero(val, DBL_TOL):
        fixXijDown(m, i, j)
    else:
        assert areEqual(val, 1.0, DBL_TOL)
        fixXijUp(m, i, j)


def fixCikl(m, i, k, l, val):
    m.Cikl[i, k, l].fix(val)
    checkCiklBounds(m, i, k, l)


def fixCi(m, i, val):
    m.Ci[i].fix(val)
    checkCiBounds(m, i)


def fixZic(m, i, c, val):
    if isZero(val, DBL_TOL):
        fixZicDown(m, i, c)
    else:
        assert areEqual(val, 1.0, DBL_TOL)
        fixZicUp(m, i, c)


def fixYikUp(m, i, k):
    if m.Yik[i, k].is_fixed():
        assert value(m.Yik[i, k]) > 0.5
        return  # unwind the stack if this has already been fixed
    for kp in m.K:
        if kp != k and m.Yik[i, kp].is_fixed() and value(m.Yik[i, kp]) > 0.5:
            raise ValueError("Tried to set multiple (k) in Yik to 1")
    m.Yik[i, k].fix(1)
    fixYiUp(m, i)
    for j in m.Ni[i]:
        if j is not None:
            for l in m.K:
                if getLB(m.Yik[j, l]) > 0.5:
                    fixXijklUp(m, i, j, k, l)
    for ip in m.I:
        if i in m.Ni[ip]:
            for l in m.K:
                if getLB(m.Yik[ip, l]) > 0.5:
                    fixXijklUp(m, ip, i, l, k)
    for kp in m.K:
        if kp != k:
            fixYikDown(m, i, kp)


def fixYiUp(m, i):
    if m.Yi[i].is_fixed():
        assert value(m.Yi[i]) > 0.5
        return
    m.Yi[i].fix(1)
    for j in m.Ni[i]:
        if j is not None:
            if getLB(m.Yi[j]) > 0.5:
                fixXijUp(m, i, j)
    for ip in m.I:
        if i in m.Ni[ip] and getLB(m.Yi[ip]) > 0.5:
            fixXijUp(m, ip, i)
    checkYikSum(m, i)


def checkYikSum(m, i):
    if len(m.Yik) == 0:
        return  # Don't introduce type-dependent variables
    sumYikLB = sum(getLB(m.Yik[i, k]) for k in m.K)
    sumYikUB = sum(getUB(m.Yik[i, k]) for k in m.K)
    assert sumYikLB <= sumYikUB
    if areEqual(sumYikUB, 1.0, 0.1) and getLB(m.Yi[i]) > 0.5:
        assert sumYikUB > 0.5
        # Can fix the one remaining Yik to one
        for k in m.K:
            if not m.Yik[i, k].is_fixed():
                fixYikUp(m, i, k)
                break
    elif sumYikUB < 0.5:
        fixYiDown(m, i)


def fixXijklUp(m, i, j, k, l, checkCikl=True):
    if m.Xijkl[i, j, k, l].is_fixed():
        assert value(m.Xijkl[i, j, k, l]) > 0.5
        return
    for kp in m.K:
        for lp in m.K:
            if (
                kp != k
                and lp != l
                and m.Xijkl[i, j, kp, lp].is_fixed()
                and value(m.Xijkl[i, j, kp, lp]) > 0.5
            ):
                raise ValueError("Tried to set multiple (k,l) in Xijkl to 1")
    m.Xijkl[i, j, k, l].fix(1)
    fixYikUp(m, i, k)
    fixYikUp(m, j, l)
    fixXijUp(m, i, j)
    for kp in m.K:
        for lp in m.K:
            if kp != k and lp != l:
                fixXijklDown(m, i, j, kp, lp)
    if checkCikl:
        checkCiklBounds(m, i, k, l)


def fixXijUp(m, i, j, checkCi=True):
    if m.Xij[i, j].is_fixed():
        assert value(m.Xij[i, j]) > 0.5
        return
    m.Xij[i, j].fix(1)
    fixYiUp(m, i)
    fixYiUp(m, j)
    if checkCi:
        checkCiBounds(m, i)


def checkCiklBounds(m, i, k, l):
    sumXijklLB = sum(
        1
        for j in m.Ni[i]
        if (
            j is not None
            and m.Xijkl[i, j, k, l].is_fixed()
            and value(m.Xijkl[i, j, k, l]) > 0.5
        )
    )
    sumXijklUB = sum(1 for j in m.Ni[i] if j is not None) - sum(
        1
        for j in m.Ni[i]
        if (
            j is not None
            and m.Xijkl[i, j, k, l].is_fixed()
            and value(m.Xijkl[i, j, k, l]) < 0.5
        )
    )
    assert sumXijklLB <= sumXijklUB
    if sumXijklUB < getLB(m.Cikl[i, k, l]):
        raise ValueError("Fixing basic vars implied an infeasible problem")
    if sumXijklLB > getUB(m.Cikl[i, k, l]):
        raise ValueError("Fixing basic vars implied an infeasible problem")
    if getUB(m.Cikl[i, k, l]) > sumXijklUB:
        m.Cikl[i, k, l].setub(sumXijklUB)
    if getLB(m.Cikl[i, k, l]) < sumXijklLB:  # Tighten Cikl bounds
        m.Cikl[i, k, l].setlb(sumXijklLB)  # Tighten Cikl bounds
    if areEqual(m.Cikl[i, k, l].lb, m.Cikl[i, k, l].ub, DBL_TOL):
        m.Cikl[i, k, l].fix(m.Cikl[i, k, l].lb)  # Fix if bounds are closed
    if areEqual(getLB(m.Cikl[i, k, l]), sumXijklUB, 0.1):
        # fix all unfixed Xijkl to one
        for j in m.Ni[i]:
            if j is not None and not m.Xijkl[i, j, k, l].is_fixed():
                fixXijklUp(m, i, j, k, l, checkCikl=False)
    if areEqual(getUB(m.Cikl[i, k, l]), sumXijklLB, 0.1):
        # fix all unfixed Xijkl to zero
        for j in m.Ni[i]:
            if j is not None and not m.Xijkl[i, j, k, l].is_fixed():
                fixXijklDown(m, i, j, k, l, checkCikl=False)


def checkCiBounds(m, i):
    sumXijLB = sum(
        1
        for j in m.Ni[i]
        if (j is not None and m.Xij[i, j].is_fixed() and value(m.Xij[i, j]) > 0.5)
    )
    sumXijUB = sum(1 for j in m.Ni[i] if j is not None) - sum(
        1
        for j in m.Ni[i]
        if (j is not None and m.Xij[i, j].is_fixed() and value(m.Xij[i, j]) < 0.5)
    )
    assert sumXijLB <= sumXijUB
    if sumXijUB < getLB(m.Ci[i]):
        raise ValueError("Fixing basic vars implied an infeasible problem")
    if sumXijLB > getUB(m.Ci[i]):
        raise ValueError("Fixing basic vars implied an infeasible problem")
    if getUB(m.Ci[i]) > sumXijUB:
        m.Ci[i].setub(sumXijUB)
    if getLB(m.Ci[i]) < sumXijLB:  # Tighten Ci bounds
        m.Ci[i].setlb(sumXijLB)  # Tighten Ci bounds
    if areEqual(m.Ci[i].lb, m.Ci[i].ub, DBL_TOL):
        m.Ci[i].fix(m.Ci[i].lb)  # Fix if bounds are closed
    if areEqual(getLB(m.Ci[i]), sumXijUB, 0.1):
        # fix all unfixed Xij to one
        for j in m.Ni[i]:
            if j is not None and not m.Xij[i, j].is_fixed():
                fixXijUp(m, i, j, checkCi=False)
    if areEqual(getUB(m.Ci[i]), sumXijLB, 0.1):
        # fix all unfixed Xij to zero
        for j in m.Ni[i]:
            if j is not None and not m.Xij[i, j].is_fixed():
                fixXijDown(m, i, j, checkCi=False)


def fixYikDown(m, i, k):
    if m.Yik[i, k].is_fixed():
        assert value(m.Yik[i, k]) < 0.5
        return
    m.Yik[i, k].fix(0)
    for j in m.Ni[i]:
        if j is not None:
            for l in m.K:
                fixXijklDown(m, i, j, k, l)
    checkYikSum(m, i)
    for l in m.K:
        fixCiklDown(m, i, k, l)


def fixYiDown(m, i):
    if m.Yi[i].is_fixed():
        assert value(m.Yi[i]) < 0.5
        return
    m.Yi[i].fix(0)
    for k in m.K:
        fixYikDown(m, i, k)
    for j in m.Ni[i]:
        if j is not None:
            fixXijDown(m, i, j)
    fixCiDown(m, i)


def fixXijklDown(m, i, j, k, l, checkCikl=True):
    if m.Xijkl[i, j, k, l].is_fixed():
        assert value(m.Xijkl[i, j, k, l]) < 0.5
        return
    m.Xijkl[i, j, k, l].fix(0)
    if checkCikl:
        checkCiklBounds(m, i, k, l)


def fixXijDown(m, i, j, checkCi=True):
    if m.Xij[i, j].is_fixed():
        assert value(m.Xij[i, j]) < 0.5
        return
    m.Xij[i, j].fix(0)
    if checkCi:
        checkCiBounds(m, i)


def fixCiklDown(m, i, k, l):
    if m.Cikl[i, k, l].is_fixed():
        assert value(m.Cikl[i, k, l]) < 0.5
        return
    m.Cikl[i, k, l].fix(0)
    checkCiklBounds(m, i, k, l)


def fixCiDown(m, i):
    if m.Ci[i].is_fixed():
        assert value(m.Ci[i]) < 0.5
        return
    m.Ci[i].fix(0)
    for k in m.K:
        for l in m.L:
            fixCiklDown(m, i, k, l)
    checkCiBounds(m, i)


def fixZicUp(m, i, c):
    if m.Zic[i, c].is_fixed():
        assert value(m.Zic[i, c]) > 0.5
        return
    m.Zic[i, c].fix(1)
    for cp in m.C:
        if cp != c:
            fixZicDown(m, i, cp)
    for l, j in enumerate(m.Ni[i]):
        if j is not None:
            for k in m.K:
                if m.Confs[c][l] == k:
                    fixYikUp(m, j, k)
                else:
                    fixYikDown(m, j, k)


def fixZicDown(m, i, c):
    if m.Zic[i, c].is_fixed():
        assert value(m.Zic[i, c]) < 0.5
        return
    m.Zic[i, c].fix(0)


def setDesignFromYik(D, m, blnSetNoneOtherwise=True):
    """Set the elements of a Design according to variables.

    Uses the Pyomo Set m.Atoms to determine which building blocks
    correspond to variable values.

    Args:
        D (Design): Design that matches the Canvas that the model
            was constructed with.
        m (ConcreteModel): Pyomo model with a solution available and an
            Atoms set defined.
        blnSetNoneOtherwise (bool): Flag to control whether the contents
            should be erased before assignment. (Default value = True)

    Returns:
        None.

    """
    if blnSetNoneOtherwise:
        D.setContents(None)
    for i, k in m.Yik.keys():
        if value(m.Yik[i, k]) > 0.5:
            D.setContent(i, k)


def setDesignFromYi(D, m, blnSetNoneOtherwise=True):
    """Set the elements of a Design according to variables.

    Uses the Pyomo Set m.Atoms to determine which building block
    corresponds to variable values. Assumes that m.Atoms[0]
    corresponds to 1 and None corresponds to 0.

    Args:
        D (Design): Design that matches the Canvas that the model
            was constructed with.
        m (ConcreteModel): Pyomo model with a solution available and an
            Atoms set defined.
        blnSetNoneOtherwise (bool): Flag to control whether the contents
            should be erased before assignment. (Default value = True)

    Returns:
        None.

    """
    if blnSetNoneOtherwise:
        D.setContents(None)
    for i in m.Yi.keys():
        if value(m.Yi[i]) > 0.5:
            D.setContent(i, m.Atoms[0])
        else:
            D.setContent(i, None)


def setDesignFromModel(D, m, blnSetNoneOtherwise=True):
    """Set the elements of a Design according to variables.

    Chooses between setDesignFromYik and setDesignFromYi.
    See those functions for more information.

    Args:
        D (Design): Design that matches the Canvas that the model
            was constructed with.
        m (ConcreteModel): Pyomo model with a solution available and an
            Atoms set defined.
        blnSetNoneOtherwise (bool): Flag to control whether the contents
            should be erased before assignment. (Default value = True)

    Returns:
        None.

    Raises:
        NotImplementedError: If it is not clear how to decide between
            competing choices.

    """
    if len(m.Yik) > 0:
        setDesignFromYik(D, m, blnSetNoneOtherwise=blnSetNoneOtherwise)
    elif len(m.Yi) > 0:
        setDesignFromYi(D, m, blnSetNoneOtherwise=blnSetNoneOtherwise)
    else:
        raise NotImplementedError(
            "User should define MaterialDescriptor Yi or Yik explicitly using valid DescriptorRule"
        )


def _validYiGivenD(m, D):
    for i in m.Yi.keys():
        if value(m.Yi[i]) > 0.5:
            if D.Contents[i] is None:
                # CASE: Atom should be present in Design but was missing
                logging.debug(
                    "_validYiGivenD value(m.Yi[{}])={} D.Contents[{}]={}".format(
                        i, value(m.Yi[i]), i, D.Contents[i]
                    )
                )
                return False
        else:
            if D.Contents[i] is not None:
                # CASE: Atom should not present in Design but was found
                logging.debug(
                    "_validYiGivenD value(m.Yi[{}])={} D.Contents[{}]={}".format(
                        i, value(m.Yi[i]), i, D.Contents[i]
                    )
                )
                return False
    return True


def _validYikGivenD(m, D):
    for i, k in m.Yik.keys():
        if value(m.Yik[i, k]) > 0.5:
            if D.Contents[i] != k:
                # CASE: Atom of type k should be present, but was not found in Design
                logging.debug(
                    "_validYikGivenD value(m.Yik[{}])={} Contents[i]={}".format(
                        (i, k), value(m.Yik[i, k]), i, D.Contents[i]
                    )
                )
                return False
        else:
            if D.Contents[i] == k:
                # CASE: Atom of type k should not be present, but was found in Design
                logging.debug(
                    "_validYikGivenD value(m.Yik[{}])={} Contents[i]={}".format(
                        (i, k), value(m.Yik[i, k]), i, D.Contents[i]
                    )
                )
                return False
    return True


def _validXijGivenD(m, D):
    for i, j in m.Xij.keys():
        if value(m.Xij[i, j]) > 0.5:
            if D.Contents[i] is None or D.Contents[j] is None:
                # CASE: Bond should be present but missing in Design
                logging.debug(
                    "_validXijGivenD value(m.Xij[{}])={} D.Contents[{}]={} D.Contents[{}]={}".format(
                        (i, j), value(m.Xij[i, j]), i, D.Contents[i], j, D.Contents[j]
                    )
                )
                return False
        else:
            if D.Contents[i] is not None and D.Contents[j] is not None:
                # CASE: No bond should be present, but was found in Design
                logging.debug(
                    "_validXijGivenD value(m.Xij[{}])={} D.Contents[{}]={} D.Contents[{}]={}".format(
                        (i, j), value(m.Xij[i, j]), i, D.Contents[i], j, D.Contents[j]
                    )
                )
                return False
    return True


def _validXijklGivenD(m, D):
    for i, j, k, l in m.Xijkl.keys():
        if value(m.Xijkl[i, j, k, l]) > 0.5:
            if D.Contents[i] != k or D.Contents[j] != l:
                # CASE: Bond of type k,l should be present, but missing in Design
                logging.debug(
                    "_validXijklGivenD value(m.Xijkl[{}])={} D.Contents[{}]={} D.Contents[{}]={}".format(
                        (i, j, k, l),
                        value(m.Xijkl[i, j, k, l]),
                        i,
                        D.Contents[{}],
                        j,
                        D.Contents[{}],
                    )
                )
                return False
        else:
            if D.Contents[i] == k and D.Contents[j] == l:
                # CASE: Bond of type k,l should not be prese,t but was found in Design
                logging.debug(
                    "_validXijklGivenD value(m.Xijkl[{}])={} D.Contents[{}]={} D.Contents[{}]={}".format(
                        (i, j, k, l),
                        value(m.Xijkl[i, j, k, l]),
                        i,
                        D.Contents[{}],
                        j,
                        D.Contents[{}],
                    )
                )
                return False
    return True


def _validCiGivenD(m, D, DBL_TOL=1e-5):
    for i in m.Ci.keys():
        Count = 0
        if D.Contents[i] is not None:
            for j in m.Ni[i]:
                if j is not None and D.Contents[j] is not None:
                    Count += 1
        if not areEqual(Count, value(m.Ci[i]), DBL_TOL):
            # CASE: Count of non-void neighbors did not match
            logging.debug("_validCiGivenD value(m.Ci[{}])={}".format(i, value(m.Ci[i])))
            return False
    return True


def _validCiklGivenD(m, D, DBL_TOL=1e-5):
    for i, k, l in m.Cikl.keys():
        Count = 0
        if D.Contents[i] == k:
            for j in m.Ni[i]:
                if j is not None and D.Contents[j] == l:
                    Count += 1
        if not areEqual(Count, value(m.Cikl[i, k, l]), DBL_TOL):
            # CASE: Count of bonds of type k,l did not mach
            logging.debug(
                "_validCiklGivenD value(m.Cikl[{}])={}".format(
                    (i, k, l), value(m.Cikl[i, k, l])
                )
            )
            return False
    return True


def _validVarsGivenD(m, D):
    if len(m.Yi) > 0 and not _validYiGivenD(m, D):
        return False
    if len(m.Yik) > 0 and not _validYikGivenD(m, D):
        return False
    if len(m.Xij) > 0 and not _validXijGivenD(m, D):
        return False
    if len(m.Xijkl) > 0 and not _validXijklGivenD(m, D):
        return False
    if len(m.Ci) > 0 and not _validCiGivenD(m, D):
        return False
    if len(m.Cikl) > 0 and not _validCiklGivenD(m, D):
        return False
    return True


def validModelSoln(m, D):
    """Assess if a Design matches variable values.

    This function checks that a building blocks in a Design
    are consistent with information from a model. Importantly,
    it is one-directional in that we do *not* check if the
    model has all of the corresponding variables present. It
    is not generally clear how to validate formulated models, but it
    is clear how to validate Designs given a model.

    Args:
        m (ConcreteModel): Pyomo model with a solution available.
        D (Design): Design with Contents set.

    Returns:
        (bool) True if Design matches the information in the model
            variables.

    """
    if not _validVarsGivenD(m, D):
        return False
    return True


# ================================================
# ==========   COMMON MODEL ELEMENTS    ==========
# ================================================
# --- COMMON CONSTRAINTS
def addConsBoundDescriptorsWithImpl(
    m,
    ConName,
    Descs,
    Impls,
    LBs=None,
    UBs=None,
    DefaultBigM=DEFAULT_BIG_M,
    blnWarnDefaultBigM=True,
):
    """Add implication constraints to a Pyomo model.

    The form of the constraint is:
    LB - M*(1-Impl) <= Desc <= UB + M*(1-Impl)

    The descriptors, implication variables, and bounds are assumed to
    be indexed over the Canvas. Big-M values are identified by
    interval arithmetic. In case this is not possible, default big-M
    values can be provided.

    Args:
        m (ConcreteModel): Pyomo model to attach constraint to.
        ConName (str): Name of new constraint to attach to the model.
        Descs (list<Expr>): Pyomo expressions to bound with implication.
        Impls (list<Var>/IndexedVar): Pyomo implication variables.
        LBs (list<float/Param>): Optional, lower bounds to enforce if
            implication is active. (Default value = None)
        UBs (list<float/Param>): Optional, upper bounds to enforce if
            implication is active. (Default value = None)
        DefaultBigM (float): Optional, big-M value to use if one can
            not be found from interval arithmetic.
            (Default value = DEFAULT_BIG_M)
        blnWarnDefaultBigM (bool): Optional, flag to enable warnings
            if the default big-M value is used. (Default value = True)

    Returns:
        None.

    """
    if LBs is not None:
        ConNameLB = ConName + "_ImplLB"
        setattr(m, ConNameLB, Constraint(m.I))
        ConLB = getattr(m, ConNameLB)
        for i in m.I:
            ExprLB = getLB(Descs[i])
            if ExprLB is None:
                if blnWarnDefaultBigM:
                    logging.warning(
                        "addConsBoundDescriptorsWithImpl used a DefaultBigM"
                    )
                MLB = -DefaultBigM
            else:
                MLB = ExprLB - LBs[i]
            ConLB.add(index=i, expr=(MLB * (1 - Impls[i]) <= Descs[i] - LBs[i]))

    if UBs is not None:
        ConNameUB = ConName + "_ImplUB"
        setattr(m, ConNameUB, Constraint(m.I))
        ConUB = getattr(m, ConNameUB)
        for i in m.I:
            ExprUB = getUB(Descs[i])
            if ExprUB is None:
                if blnWarnDefaultBigM:
                    logging.warning(
                        "addConsBoundDescriptorsWithImpl used a DefaultBigM"
                    )
                MUB = DefaultBigM
            else:
                MUB = ExprUB - UBs[i]
            ConUB.add(index=i, expr=(MUB * (1 - Impls[i]) >= Descs[i] - UBs[i]))


def addConsIndFromDescriptors(
    m,
    ConName,
    Descs,
    Inds,
    LBs=None,
    UBs=None,
    Eps=DEFAULT_EPS,
    DefaultBigM=DEFAULT_BIG_M,
    blnWarnDefaultBigM=True,
):
    """Add indicator constraints to a Pyomo model.

    The form of the constraint are:
        Desc >= UB + Eps - M*(UBInd)
        Desc <= LB - Eps + M*(LBInd)
        Ind >= LBInd + UBInd - 1

    The descriptors, indication variables, and bounds are assumed to
    be indexed over the Canvas. Big-M values are identified by
    interval arithmetic. In case this is not possible, default big-M
    values can be provided.

    Args:
        m (ConcreteModel): Pyomo model to attach constraint to.
        ConName (str): Name of new constraint to attach to the model.
        Descs (list<Expr>): Pyomo expressions to bound with implication.
        Inds (list<Var>/IndexedVar): Pyomo indicator variables.
        LBs (list<float/Param>): Optional, lower bounds to enforce if
            implication is active. (Default value = None)
        UBs (list<float/Param>): Optional, upper bounds to enforce if
            implication is active. (Default value = None)
        Eps (float): Optional, tolerance to use for modeling
            constraint equality.
        DefaultBigM (float): Optional, big-M value to use if one can
            not be found from interval arithmetic.
            (Default value = DEFAULT_BIG_M)
        blnWarnDefaultBigM (bool): Optional, flag to enable warnings
            if the default big-M value is used. (Default value = True)

    Returns:
        None.

    """
    if LBs is not None and UBs is not None:
        ConNameLink = ConName + "_Link"
        VarNameLBInd = ConName + "_LBIndi"
        VarNameUBInd = ConName + "_UBIndi"
        setattr(m, ConNameLink, Constraint(m.I))
        setattr(m, VarNameLBInd, Var(m.I, bounds=Binary, dense=False))
        setattr(m, VarNameUBInd, Var(m.I, bounds=Binary, dense=False))
        ConLink = getattr(m, ConNameLink)
        IndsLB = getattr(m, VarNameLBInd)
        IndsUB = getattr(m, VarNameUBInd)
        for i in m.I:
            ConLink.add(index=i, expr=(Inds[i] >= IndsLB[i] + IndsUB[i] - 1))
    elif LBs is not None:
        IndsLB = Inds
    elif UBs is not None:
        IndsUB = Inds
    else:
        return

    if LBs is not None:
        ConNameLB = ConName + "_LB"
        setattr(m, ConNameLB, Constraint(m.I))
        ConLB = getattr(m, ConNameLB)
        for i in m.I:
            ExprUB = getUB(Descs[i])
            if ExprUB is None:
                if blnWarnDefaultBigM:
                    logging.warning("addConsIndFromDescriptors used a DefaultBigM")
                MLB = DefaultBigM
            else:
                MLB = -LBs[i] + Eps + ExprUB
            ConLB.add(index=i, expr=(Descs[i] <= LBs[i] - Eps + MLB * (IndsLB[i])))
    if UBs is not None:
        ConNameUB = ConName + "_UB"
        setattr(m, ConNameUB, Constraint(m.I))
        ConUB = getattr(m, ConNameUB)
        for i in m.I:
            ExprLB = getLB(Descs[i])
            if ExprLB is None:
                if blnWarnDefaultBigM:
                    logging.warning("addConsIndFromDescriptors used a DefaultBigM")
                MUB = -DefaultBigM
            else:
                MUB = -UBs[i] - Eps + ExprLB
            ConUB.add(index=i, expr=(Descs[i] >= UBs[i] + Eps + MUB * (IndsUB[i])))


def addConsLocalBudgets(m, ConName, Vs, Subsets=None, BudgetLBs=None, BudgetUBs=None):
    """Add budget constraints to a Pyomo model.

    The form of the constraints are:
        BudgetLB <= sum(Vs[j] for j in Subsets[i]) <= BudgetUB

    The variables, subsets, and budget bounds are assumed to be
    indexed over Canvas locations. The budget is written over
    local neighborhoods defined in Subsets. If the Subsets keyword
    is not provided, then the Canvas neighborhoods are used.

    Args:
        m (ConcreteModel): Pyomo model to attach constraint to.
        ConName (str): Name of new constraint to attach to the model.
        Vs (list<Var/Expr>): Pyomo expressions to add up in a local budget.
        Subsets: Optional, Neighborhoods to sum over. If not provided,
            the Canvas Neighborhoods are used. (Default value = None)
        BudgetLBs: Optional, lower bounds on the local summations to
            enforce. (Default value = None)
        BudgetUBs: Optional, upper bounds on the local summations to
            enforce. (Default value = None)

    Returns:
        None.

    """
    if BudgetLBs is None and BudgetUBs is None:
        return
    if BudgetLBs is None:
        BudgetLBs = [None] * len(m.I)
    elif BudgetUBs is None:
        BudgetUBs = [None] * len(m.I)
    if Subsets is None:
        Subsets = m.Ni
    else:
        assert len(Subsets) == len(m.I)
    setattr(m, ConName, Constraint(m.I))
    Con = getattr(m, ConName)
    for i in m.I:
        if Subsets[i] is not None:
            Expr = 0
            for j in Subsets[i]:
                if j is not None:
                    Expr += Vs[j]
            Con.add(index=i, expr=(BudgetLBs[i], Expr, BudgetUBs[i]))


def addConGlobalBudget(m, ConName, Vs, Subset=None, BudgetLB=None, BudgetUB=None):
    """Add budget constraints to a Pyomo model.

    The form of the constraint is:
        BudgetLB <= sum(Vs[i] for i in Subset) <= BudgetUB

    The variables are assumed to be indexed over Canvas locations.
    In contrast to addConLocalBudgets, this function applies a single
    budget constraint written over the entire Canvas. If the Subsets
    keyword is provided, then a subset of Canvas locations are added.

    Args:
        m (ConcreteModel): Pyomo model to attach constraint to.
        ConName (str): Name of new constraint to attach to the model.
        Vs (list<Var/Expr>): Pyomo expressions to add up in a budget.
        Subset: Optional, indices to sum over. If not provided, the
            Canvas indices are used. (Default value = None)
        BudgetLB: Optional, lower bound on the summation to enforce.
            (Default value = None)
        BudgetUB: Optional, upper bound on the summation to enforce.
            (Default value = None)

    Returns:
        None.

    """
    if BudgetLB is None and BudgetUB is None:
        return
    if Subset is None:
        Subset = m.I
    Expr = sum(Vs[i] for i in Subset)
    setattr(m, ConName, Constraint(expr=(BudgetLB, Expr, BudgetUB)))


def addConsLocalBudgetsByTypes(
    m, ConName, Vs, Subsets=None, Types=None, BudgetLBs=None, BudgetUBs=None
):
    """Add budget constraints to a Pyomo model.

    The form of the constraints are:
        BudgetLBs[k] <= sum(Vs[j,k] for j in Subsets[i]) <= BudgetUBs[k]

    The variables, subsets, and budget bounds are assumed to be
    indexed over Canvas locations. The budget is written over
    local neighborhoods defined in Subsets. If the Subsets keyword
    is not provided, then the Canvas neighborhoods are used.

    Args:
        m (ConcreteModel): Pyomo model to attach constraint to.
        ConName (str): Name of new constraint to attach to the model.
        Vs (dict<Var/Expr>): Pyomo expressions to add up in a local budget.
        Subsets: Optional, Neighborhoods to sum over. If not provided,
            the Canvas Neighborhoods are used. (Default value = None)
        BudgetLBs: Optional, lower bounds on the local summations to
            enforce. (Default value = None)
        BudgetUBs: Optional, upper bounds on the local summations to
            enforce. (Default value = None)

    Returns:
        None.

    """
    if BudgetLBs is None and BudgetUBs is None:
        return
    if Subsets is None:
        Subsets = m.Ni
    else:
        assert len(Subsets) == len(m.I)
    if Types is None:
        Types = m.K
    if BudgetLBs is None:
        BudgetLBs = {(i, k): None for i in m.I for k in Types}
    elif BudgetUBs is None:
        BudgetUBs = {(i, k): None for i in m.I for k in Types}
    setattr(m, ConName, Constraint(m.I, Types))
    Con = getattr(m, ConName)
    for k in Types:
        for i in m.I:
            if Subsets[i] is not None:
                Expr = 0
                for j in Subsets[i]:
                    if j is not None:
                        Expr += Vs[j, k]
                Con.add(index=(i, k), expr=(BudgetLBs[i, k], Expr, BudgetUBs[i, k]))


def addConsGlobalBudgetsByTypes(
    m, ConName, Vs, Subset=None, Types=None, BudgetLBs=None, BudgetUBs=None
):
    """Add budget constraints to a Pyomo model.

    The form of the constraint is:
        BudgetLBs[k] <= sum(Vs[i,k] for i in Subset) <= BudgetUBs[k]

    The variables are assumed to be indexed over Canvas locations
    and types.
    In contrast to addConLocalBudgetsByTypes, this function applies
    a single budget constraint written over the entire Canvas. If the
    Subsets keyword is provided, then a subset of Canvas locations
    are added.

    Args:
        m (ConcreteModel): Pyomo model to attach constraint to.
        ConName (str): Name of new constraint to attach to the model.
        Vs (dict<k:Var/Expr>): Pyomo expressions to add up in a budget.
        Subset: Optional, indices to sum over. If not provided, the
            Canvas indices are used. (Default value = None)
        BudgetLBs: Optional, lower bound on the summation to enforce.
            (Default value = None)
        BudgetUBs: Optional, upper bound on the summation to enforce.
            (Default value = None)

    Returns:
        None.

    """
    if BudgetLBs is None and BudgetUBs is None:
        return
    if Subset is None:
        Subset = m.I
    if Types is None:
        Types = m.K
    if BudgetLBs is None:
        BudgetLBs = {k: None for k in Types}
    elif BudgetUBs is None:
        BudgetUBs = {k: None for k in Types}
    setattr(m, ConName, Constraint(Types))
    Con = getattr(m, ConName)
    for k in Types:
        Expr = sum(Vs[i, k] for i in Subset)
        Con.add(index=k, expr=(BudgetLBs[k], Expr, BudgetUBs[k]))


def addConsZicMutExc(m):
    """Add constraints for mutually exclusive conformations.

    Args:
        m (ConcreteModel): Pyomo model to attach constraints to.

    Returns:
        None.

    """

    def _ruleZicMutExc(m, i):
        return sum(m.Zic[i, c] for c in m.C) <= 1

    m.AssignZicMutExc = Constraint(m.I, rule=_ruleZicMutExc)


def addConsZicColExh(m):
    """
    Add constraints for collectively exhaustive conformations.

    Args:
        m (ConcreteModel): Pyomo model to attach constraints to.

    Returns:
        None.

    """

    def _ruleZicColExh(m, i):
        return sum(m.Zic[i, c] for c in m.C) >= 1

    m.AssignZicColExh = Constraint(m.I, rule=_ruleZicColExh)


def addConsZicMutExcColExh(m):
    """
    Add constraints for mutually exclusive, collectively
    exhaustive of conformations.

    Args:
        m (ConcreteModel): Pyomo model to attach constraints to.

    Returns:
        None.

    """

    def _ruleZicMutExcColExh(m, i):
        return sum(m.Zic[i, c] for c in m.C) == 1

    m.AssignZicMutExcColExh = Constraint(m.I, rule=_ruleZicMutExcColExh)


def addConsZicFromYi(m, blnConfsAreMutExc=True, blnConfsAreColExh=False):
    """
    Add constraints for indicating conformations.

    This function produces indicator constraints for the presence of
    conformation 'c' at location 'i' if building blocks in neighboring
    locations match the pattern for conformation 'c'. Neighboring
    locations are defined by the set m.Ni. In this function, any
    types of building block are considered. The variable Yi should be
    equal to 1 when any building block is present.
    This function only writes assignment constraints for m.Zic variables
    previously referenced in the model.

    Args:
        m (ConcreteModel): Pyomo model to attach constraints to.
        blnConfsAreMutExc (bool): Flag to indicate whether to add
            constraints enforcing mutually exclusive conformations.
            (Default value = True)
        blnConfsAreColExh (bool): Flag to indicate whether to add
            constraints enforcing collectively exhaustive conformations.
            NOTE: Since this function only loops over Zic variables
            previously referenced in the model, it is important
            to be carefule when using this flag. (Default value = False)

    Returns:
        None.

    """
    m.AssignZicFromYi1 = Constraint(m.I, m.C, m.I)
    m.AssignZicFromYi2 = Constraint(m.I, m.C)
    for i, c in m.Zic.keys():
        assert len(m.Ni[i]) == len(m.Confs[c])
        RelaxExpr = 0
        for l, j in enumerate(m.Ni[i]):
            if j is not None:
                if m.Confs[c][l] is not None:
                    RelaxExpr += m.Yi[j] - 1
                    m.AssignZicFromYi1.add(
                        index=(i, c, j), expr=(m.Zic[i, c] <= m.Yi[j])
                    )
                else:
                    assert m.Confs[c][l] is None
                    RelaxExpr += -m.Yi[j]
                    m.AssignZicFromYi1.add(
                        index=(i, c, j), expr=(m.Zic[i, c] <= 1 - m.Yi[j])
                    )
        m.AssignZicFromYi2.add(index=(i, c), expr=(m.Zic[i, c] >= 1 + RelaxExpr))
    if blnConfsAreMutExc and blnConfsAreColExh:
        addConsZicMutExcColExh(m)
    elif blnConfsAreMutExc:
        addConsZicMutExc(m)
    elif blnConfsAreColExh:
        addConsZicColExh(m)


def addConsZicFromYiLifted(m, blnConfsAreMutExc=True, blnConfsAreColExh=False):
    """Add constraints for indicating conformations.

    See documentation for addConsZicFromYi for more information.
    This version uses a lifted set of constraints.

    Args:
        m (ConcreteModel): Pyomo model to attach constraints to.
        blnConfsAreMutExc (bool): Flag to indicate whether to add
            constraints enforcing mutually exclusive conformations.
            (Default value = True)
        blnConfsAreColExh (bool): Flag to indicate whether to add
            constraints enforcing collectively exhaustive conformations.
            NOTE: Since this function only loops over Zic variables
            previously referenced in the model, it is important
            to be carefule when using this flag. (Default value = False)

    Returns:
        None.

    """
    m.AssignZicFromYiLifted1 = Constraint(m.I, m.I)
    m.AssignZicFromYiLifted2 = Constraint(m.I, m.I)
    m.AssignZicFromYiLifted3 = Constraint(m.I, m.C)
    iAlreadyProcessed = set()
    for i, _ in m.Zic.keys():
        if i not in iAlreadyProcessed:
            for l, j in enumerate(m.Ni[i]):
                if j is not None:
                    PosZic = sum(
                        m.Zic[i, c]
                        for c in m.Zic[i, :].wildcard_keys()
                        if m.Confs[c][l] is not None
                    )
                    NegZic = sum(
                        m.Zic[i, c]
                        for c in m.Zic[i, :].wildcard_keys()
                        if m.Confs[c][l] is None
                    )
                    if PosZic != 0:
                        m.AssignZicFromYiLifted1.add(
                            index=(i, j), expr=(PosZic <= m.Yi[j])
                        )
                    if NegZic != 0:
                        m.AssignZicFromYiLifted2.add(
                            index=(i, j), expr=(NegZic <= 1 - m.Yi[j])
                        )
            iAlreadyProcessed.add(i)
    for i, c in m.Zic.keys():
        RelaxExpr = 0
        for l, j in enumerate(m.Ni[i]):
            if j is not None:
                if m.Confs[c][l] is not None:
                    RelaxExpr += m.Yi[j] - 1
                else:
                    RelaxExpr += -m.Yi[j]
        m.AssignZicFromYiLifted3.add(index=(i, c), expr=(m.Zic[i, c] >= 1 + RelaxExpr))
    if blnConfsAreMutExc and blnConfsAreColExh:
        addConsZicMutExcColExh(m)
    elif blnConfsAreMutExc:
        addConsZicMutExc(m)
    elif blnConfsAreColExh:
        addConsZicColExh(m)


def addConsZicFromYik(m, blnConfsAreMutExc=True, blnConfsAreColExh=False):
    """Add constraints for indicating conformations.

    This function produces indicator constraints for the presence of
    conformation 'c' at location 'i' if building blocks of type 'k'
    in neighboring locations match the pattern for conformation 'c'.
    Neighboring locations are defined by the set m.Ni. Building block
    types are defined in m.Atoms and correspond to the 'k' index.
    The variable Yik should be equal to 1 when a building block of type
    k is present in the conformation.
    This function only writes assignment constraints for m.Zic variables
    previously referenced in the model.

    Args:
        m (ConcreteModel): Pyomo model to attach constraints to.
        blnConfsAreMutExc (bool): Flag to indicate whether to add
            constraints enforcing mutually exclusive conformations.
            (Default value = True)
        blnConfsAreColExh (bool): Flag to indicate whether to add
            constraints enforcing collectively exhaustive conformations.
            NOTE: Since this function only loops over Zic variables
            previously referenced in the model, it is important
            to be carefule when using this flag. (Default value = False)

    Returns:
        None.

    """
    m.AssignZicFromYik1 = Constraint(m.I, m.C, m.I, m.K)
    m.AssignZicFromYik2 = Constraint(m.I, m.C)
    for i, c in m.Zic.keys():
        assert len(m.Ni[i]) == len(m.Confs[c])
        RelaxExpr = 0
        for l, j in enumerate(m.Ni[i]):
            if j is not None:
                if m.Confs[c][l] is None:
                    RelaxExpr += m.Yi[j]
                for k in m.K:
                    if m.Confs[c][l] == k:
                        RelaxExpr += 1 - m.Yik[j, k]
                        m.AssignZicFromYik1.add(
                            index=(i, c, j, k), expr=(m.Zic[i, c] <= m.Yik[j, k])
                        )
                    else:
                        m.AssignZicFromYik1.add(
                            index=(i, c, j, k), expr=(m.Zic[i, c] <= 1 - m.Yik[j, k])
                        )
        m.AssignZicFromYik2.add(index=(i, c), expr=(m.Zic[i, c] >= 1 - RelaxExpr))
    if blnConfsAreMutExc and blnConfsAreColExh:
        addConsZicMutExcColExh(m)
    elif blnConfsAreMutExc:
        addConsZicMutExc(m)
    elif blnConfsAreColExh:
        addConsZicColExh(m)


def addConsZicFromYikLifted(m, blnConfsAreMutExc=True, blnConfsAreColExh=False):
    """Add constraints for indicating conformations.

    See documentation for addConsZicFromYik for more information.
    This version uses a lifted set of constraints.

    Args:
        m (ConcreteModel): Pyomo model to attach constraints to.
        blnConfsAreMutExc (bool): Flag to indicate whether to add
            constraints enforcing mutually exclusive conformations.
            (Default value = True)
        blnConfsAreColExh (bool): Flag to indicate whether to add
            constraints enforcing collectively exhaustive conformations.
            NOTE: Since this function only loops over Zic variables
            previously referenced in the model, it is important
            to be carefule when using this flag. (Default value = False)

    Returns:
        None.

    """
    m.AssignZicFromYikLifted1 = Constraint(m.I, m.I, m.K)
    m.AssignZicFromYikLifted2 = Constraint(m.I, m.I, m.K)
    m.AssignZicFromYikLifted3 = Constraint(m.I, m.C)
    iAlreadyProcessed = set()
    for i, _ in m.Zic.keys():
        if i not in iAlreadyProcessed:
            for l, j in enumerate(m.Ni[i]):
                if j is not None:
                    for k in m.K:
                        PosZic = sum(
                            m.Zic[i, c]
                            for c in m.Zic[i, :].wildcard_keys()
                            if m.Confs[c][l] == k
                        )
                        NegZic = sum(
                            m.Zic[i, c]
                            for c in m.Zic[i, :].wildcard_keys()
                            if m.Confs[c][l] != k
                        )
                        if PosZic != 0:
                            m.AssignZicFromYikLifted1.add(
                                index=(i, j, k), expr=(PosZic <= m.Yik[j, k])
                            )
                        if NegZic != 0:
                            m.AssignZicFromYikLifted2.add(
                                index=(i, j, k), expr=(NegZic <= 1 - m.Yik[j, k])
                            )
            iAlreadyProcessed.add(i)
    for i, c in m.Zic.keys():
        RelaxExpr = 0
        for l, j in enumerate(m.Ni[i]):
            if j is not None:
                if m.Confs[c][l] is None:
                    RelaxExpr += m.Yi[j]
                for k in m.K:
                    if m.Confs[c][l] == k:
                        RelaxExpr += 1 - m.Yik[j, k]
        m.AssignZicFromYikLifted3.add(index=(i, c), expr=(m.Zic[i, c] >= 1 - RelaxExpr))
    if blnConfsAreMutExc and blnConfsAreColExh:
        addConsZicMutExcColExh(m)
    elif blnConfsAreMutExc:
        addConsZicMutExc(m)
    elif blnConfsAreColExh:
        addConsZicColExh(m)


# --- COMMON OBJECTIVE FUNCTIONS
def addObjMaxSumZi(m):
    """Add objective for maximizing target sites.

    Args:
        m (ConcreteModel): Pyomo model to add objective to.

    Returns:
        None.

    """
    m.Obj = Objective(expr=sum(m.Zi[i] for i in m.I), sense=maximize)


def addObjMinSumYi(m):
    """Add objective for minimizing building blocks used.

    This objective is mainly useful for debugging models.
    Fix some (supposed) feasible solution, use this objective,
    and see which building blocks were realized.

    Args:
        m (ConcreteModel): Pyomo model to add objective to.

    Returns:
        None.

    """
    m.Obj = Objective(expr=sum(m.Yi[i] for i in m.I), sense=minimize)


def addObjMaxSumZic(m, Alphas=None, ISubset=None, ConfSubset=None):
    """Add objective for maximizing target conformations.

    Args:
        m (ConcreteModel): Pyomo model to add objective to.
        Alphas (list<float/Param>): Optional, weighting to give each
            conformation. If not present, assumes a weight of 1.0.
            (Default value = None)
        ISubset (list<int>/Set): Optional, subset of locations to
            consider. If not present, assumes the entire Canvas.
            (Default value = None)
        ConfSubset: Optional, subset of conformations to consider.
            If not present, assumes all conformations.
            (Default value = None)

    Returns:
        None.

    """
    if ISubset is None:
        ISubset = m.I
    if ConfSubset is None:
        ConfSubset = m.C
    if Alphas is None:
        Alphas = [1] * len(ISubset)
    m.Obj = Objective(
        expr=sum(Alphas[c] * m.Zic[i, c] for i in ISubset for c in ConfSubset),
        sense=maximize,
    )
